import muteria.common.mix as common_mix
ERROR_HANDLER = common_mix.ErrorHandler

from muteria.drivers.testgeneration.tools_by_languages.c.klee.driver_config \
                                                        import DriverConfigKlee

class DriverConfigShadow (DriverConfigKlee):
    def __init__(self, keep_first_test=False, gen_timeout_is_per_test=False, \
                                                                    **kwargs):
        DriverConfigKlee.__init__(self, **kwargs)
        ERROR_HANDLER.assert_true(type(keep_first_test) == bool, \
                        "invalid keep_first_test type. Must be bool", __file__)
        self.keep_first_test = keep_first_test
        self.gen_timeout_is_per_test = gen_timeout_is_per_test
    #~ def __init__()

    def get_keep_first_test(self):
        return self.keep_first_test
    #~ def get_keep_first_test()
    
    def get_gen_timeout_is_per_test(self):
        return self.gen_timeout_is_per_test
    #~ def get_gen_timeout_is_per_test()
#~ class DriverConfigShadow



import os
import sys
import imp
from distutils.spawn import find_executable
import logging
import struct
import shutil
import glob

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

ERROR_HANDLER = common_mix.ErrorHandler

from muteria.drivers import DriversUtils

from muteria.drivers.testgeneration.testcase_formats.ktest.ktest import \
                                                                KTestTestFormat

class Misc:
    @staticmethod
    def import_ktest_tool (custom_binary_dir=None):
        if custom_binary_dir is None:
            ktt_full_path = find_executable('ktest-tool')
            ERROR_HANDLER.assert_true(ktt_full_path is not None, \
                            "ktest-tool is not present on the PATH", __file__)
        else:
            ERROR_HANDLER.assert_true(os.path.isdir(custom_binary_dir), \
                            "Custom_binary_dir inexistant ({})".format(\
                                            custom_binary_dir), __file__)
            ktt_full_path = os.path.join(custom_binary_dir, 'ktest-tool')
            ERROR_HANDLER.assert_true(os.path.isfile(ktt_full_path), \
                        "ktest-tool not found in custom binary dir", __file__)
        ktest_tool = imp.load_source("ktest-tool", ktt_full_path)

        return ktest_tool
    #~ def import_ktest_tool ()
    
    @staticmethod
    def get_must_exist_dirs_of_ktests(ktest_key_to_file, custom_binary_dir=None):
        """ Takes a dict of ktest key to ktest file
            returns the dict of ktest key to list of required dirs
        """
        ktest_tool = Misc.import_ktest_tool(\
                                        custom_binary_dir=custom_binary_dir)
        ktest2reqdir = {}
        for ktest_key, ktest_file in ktest_key_to_file.items():
            # Compute the list of req_dirs
            req_dir_set = set()
            b = ktest_tool.KTest.fromfile(ktest_file)
            prev_name = None
            for ind, obj in enumerate(b.objects):
                name, data = obj
                if prev_name is not None and \
                            name == prev_name + b'-stat' and len(data) == 144:
                    # case of file stat
                    prev_name = None
                    continue
                
                prev_name = name
                
                # is it a path?
                d, f = os.path.split(os.path.normpath(name))
                if len(d) > 0 and len(f) > 0:
                    if not os.path.isabs(d):
                        ERROR_HANDLER.assert_true(b'..' not in d, \
                                "acces to parent of cur dir not allowed. " + \
                                    "ktest is {}".format(ktest_file), __file__)
                        req_dir_set.add(d.decode('UTF-8', 'backslashreplace'))
                        ERROR_HANDLER.assert_true((ind + 1) < len(b.objects), \
                              "invalid ktest, '-stat' missing " +\
                                "(ind is {}). ktest file is {}".format(\
                                                       ind, ktest_file), __file__)
                        ERROR_HANDLER.assert_true(\
                                      b.objects[ind+1][0] == (name + b"-stat"), \
                              "Invalid ktest ofr bug? ({})".format(ktest_file), \
                                                                        __file__)
                
            ktest2reqdir[ktest_key] = list(req_dir_set)
        return ktest2reqdir
    #~ def get_must_exist_dirs_of_ktests()
#~ class Misc

class FileShortNames:
    def __init__(self):
        #schar = [chr(i) for i in range(ord('A'),ord('Z')+1)\
        #                                    +range(ord('a'),ord('z')+1)]
        #ShortNames = {'names':[z[0]+z[1] for z in \
        #                    itertools.product(schar,['']+schar)], 'pos':0}

        # From KLEE: 'void klee_init_fds()' in runtime/POSIX/fd_init.c
        self.ShortNames = [bytes(chr(i), 'utf-8') for i in range(ord('A'), 256)]  
        self.pos = 0
    def reinitialize_count (self):
        self.pos = 0
    def getShortname(self): 
        self.pos += 1
        if self.pos >= len(self.ShortNames):
            ERROR_HANDLER.error_exit("too many file arguments, "
                                        "exeeded shortname list", __file__)
        return self.ShortNames[self.pos-1]
#~ class FileShortNames:

class ConvertCollectKtestsSeeds:
    tar_gz = ".tar.gz"
    test2semudirMapFile = "test2semudirMapFile.json"

    def __init__(self, custom_binary_dir=None):
        self.ktest_tool = Misc.import_ktest_tool(\
                                        custom_binary_dir=custom_binary_dir)
    #~ def __init__ ()

    def generate_seeds_from_various_ktests (self, dest_dir, \
                                            src_old_shadow_zesti_ktest_dir, \
                                   src_new_klee_ktest_dir_or_sym_args=None, \
                                            klee_ktest_is_sym_args=False, \
                                      compress_dest=True, skip_failure=False):
        """
        """

        ERROR_HANDLER.assert_true(os.path.isdir(os.path.dirname(dest_dir)), \
                                "dest dir parent does not exists", __file__)
        ERROR_HANDLER.assert_true(not os.path.isdir(dest_dir), \
                                        "dest dir already exists", __file__)
        ERROR_HANDLER.assert_true(not (compress_dest and \
                                        os.path.isfile(dest_dir+self.tar_gz)),\
                                "dest dir compressed already exists", __file__)
        
        if klee_ktest_is_sym_args:
            src_new_klee_ktest_dir = None
            klee_ktest_sym_args = src_new_klee_ktest_dir_or_sym_args
        else:
            klee_ktest_sym_args = None
            src_new_klee_ktest_dir = src_new_klee_ktest_dir_or_sym_args
            
        os.mkdir(dest_dir)

        tmpdir = os.path.join(dest_dir, '.tmp')
        os.mkdir(tmpdir)

        ERROR_HANDLER.assert_true(src_old_shadow_zesti_ktest_dir is not None \
                                    or src_new_klee_ktest_dir is not None, \
                                    "at least one src must be set", __file__)

        # decompress shadow_zesti ktest dir if compressed
        if src_old_shadow_zesti_ktest_dir is not None and \
                        src_old_shadow_zesti_ktest_dir.endswith(self.tar_gz):
            # decompress
            tmp_ar = src_old_shadow_zesti_ktest_dir
            src_old_shadow_zesti_ktest_dir = \
                            src_old_shadow_zesti_ktest_dir[:-len(self.tar_gz)]
            common_fs.TarGz.decompressDir(tmp_ar, \
                            os.path.dirname(src_old_shadow_zesti_ktest_dir), \
                                                    remove_in_archive=False)

        # decompress new_klee ktest dir if compressed
        if src_new_klee_ktest_dir is not None and \
                                src_new_klee_ktest_dir.endswith(self.tar_gz):
            # decompress
            tmp_ar = src_new_klee_ktest_dir
            src_new_klee_ktest_dir = src_new_klee_ktest_dir[:-len(self.tar_gz)]
            common_fs.TarGz.decompressDir(tmp_ar, \
                                    os.path.dirname(src_new_klee_ktest_dir), \
                                                    remove_in_archive=False)

        # Get test to zesti dir map (from the src_old... dir) TODO
        test2zestidirMap = {}
        for tc_name in os.listdir(src_old_shadow_zesti_ktest_dir):
            tc_name_dir = os.path.join(src_old_shadow_zesti_ktest_dir, tc_name)
            if os.path.isdir(tc_name_dir):
                for kleeout in os.listdir(tc_name_dir):
                    kleeout_dir = os.path.join(tc_name_dir, kleeout)
                    if os.path.isdir(kleeout_dir) \
                                and len(glob.glob(os.path.join(kleeout_dir, \
                                    "*"+KTestTestFormat.ktest_extension))) > 0:
                        test2zestidirMap[tc_name] = kleeout_dir

        # Get the seeds
        # Get sym args, convert and merge
        zest_sym_args_param = None
        zestKTContains = None
        klee_sym_args_param = None
        kleeKTContains = None

        zestKtests = []
        for tc, tcdir in test2zestidirMap.items():
            listKtestFiles = glob.glob(os.path.join(tcdir, \
                                        "*"+KTestTestFormat.ktest_extension))
            ERROR_HANDLER.assert_true(len(listKtestFiles) == 1, \
                            "Error: more than 1 or no ktest from Zesti" + \
                            " for tests: "+tc+", zestiout: "+tcdir, __file__)
            for ktestfile in listKtestFiles:
                zestKtests.append(ktestfile)
        # refactor the ktest fom zesti 
        zest_sym_args_param, zestKTContains = self._getSymArgsFromZestiKtests(\
                                                zestKtests, test2zestidirMap, \
                                                     skip_failure=skip_failure)

        # get new klee stuffs
        if src_new_klee_ktest_dir is not None:
            new_klee_test_list = []
            cwd = os.getcwd()
            os.chdir(src_new_klee_ktest_dir)
            for root, _, files in os.walk('.'):
                for f in files:
                    tc = os.path.normpath(os.path.join(root, f))
                    if tc.endswith(KTestTestFormat.ktest_extension):
                        new_klee_test_list.append(tc)
            os.chdir(cwd)
            klee_sym_args_param, kleeKTContains = \
                                    self._loadAndGetSymArgsFromKleeKTests (\
                                                    new_klee_test_list, \
                                                    src_new_klee_ktest_dir)
        elif klee_ktest_sym_args is not None:
            _, kleeKTContains = \
                            self._loadAndGetSymArgsFromKleeKTests([], None)
            klee_sym_args_param = [" ".join(l) for l in klee_ktest_sym_args]
        else:
            klee_sym_args_param, kleeKTContains = None, None
            
        sym_args_param, test2semudirMap = self._mergeZestiAndKleeKTests (\
                                        dest_dir, \
                                        zestKTContains, zest_sym_args_param, \
                                        kleeKTContains, klee_sym_args_param)
        # convert sym_args_param from this format: 
        # ["-sym-arg 1", "-sym-args 2 3 4", ...]
        # into this format:
        # [["-sym-arg", "1"], ["-sym-args", "2", "3", "4"], ...]
        sym_args_param = [g.split() for g in sym_args_param]
        common_fs.dumpJSON([sym_args_param, \
                        self._stripRootTest2Dir(dest_dir, test2semudirMap)], \
                            os.path.join(dest_dir, self.test2semudirMapFile))

        # delete tmpdir
        shutil.rmtree(tmpdir)

        # Fdupes the seedDir.
        cmd_ret, out, _ = DriversUtils.execute_and_get_retcode_out_err(\
                                        prog='fdupes',
                                        args_list=['-r', '-d', '-N', dest_dir],
                            out_on=True, err_on=True, merge_err_to_out=True)
        if cmd_ret != 0:
            ERROR_HANDLER.error_exit ("fdupes failed on SeedDir:\n{}".format(\
                                                                out), __file__)

        # compress destdir
        if compress_dest:
            common_fs.TarGz.compressDir(dest_dir, remove_in_directory=True)
    #~ def generate_seeds_from_various_ktests()

    def get_ktests_sym_args(self, ktests_dir, compressed=True):
        """ get sym args, if not there, create from ktests
        """
        if compressed:
            tmp_ar = ktests_dir
            ktests_dir = ktests_dir[:-len(self.tar_gz)]
            ERROR_HANDLER.assert_true(not os.path.isdir(ktests_dir), \
                    "make sure the uncompressed dir is abscent ({})".format(\
                                                        ktests_dir), __file__)
            common_fs.TarGz.decompressDir(tmp_ar, \
                                                os.path.dirname(ktests_dir), \
                                                    remove_in_archive=False)

        ERROR_HANDLER.assert_true(os.path.isdir(ktests_dir), \
                        "ktests dir {} is issing".format(ktests_dir), __file__)

        datfile = os.path.join(ktests_dir, self.test2semudirMapFile)
        ERROR_HANDLER.assert_true(os.path.isfile(datfile), \
                    "datfile {} is missing".format(self.test2semudirMapFile), \
                                                                    __file__)
        
        dat = common_fs.loadJSON(datfile)

        if compressed:
            shutil.rmtree(ktests_dir)

        return dat[0]
    #~ def get_ktests_sym_args()

    ##################################
    ########### PRIVATE ##############
    ##################################

    @staticmethod
    def _stripRootTest2Dir (rootdir, test2dir):
        res = {}
        for tc in test2dir:
            res[tc] = os.path.relpath(test2dir[tc], rootdir)
        return res
    #~ def _stripRootTest2Dir ()

    @staticmethod
    def _prependRootTest2Dir (rootdir, test2dir):
        res = {}
        for tc in test2dir:
            res[tc] = os.path.join(rootdir, test2dir[tc])
        return res
    #~ def _prependRootTest2Dir ()

    @staticmethod
    def _ktestToFile(ktestData, outfilename):
        """
            serialize the ktest data(ktest_tool.KTest type) into a ktest file
        """
        with open(outfilename, "wb") as f:
            f.write(b'KTEST')
            f.write(struct.pack('>i', ktestData.version))
            
            f.write(struct.pack('>i', len(ktestData.args)))
            for i in range(len(ktestData.args)):
                f.write(struct.pack('>i', len(ktestData.args[i])))
                f.write(ktestData.args[i].encode(encoding="ascii"))
            
            if ktestData.version > 2:
                f.write(struct.pack('>i', ktestData.symArgvs))
                f.write(struct.pack('>i', ktestData.symArgvLen))
            
            f.write(struct.pack('>i', len(ktestData.objects)))
            for i in range(len(ktestData.objects)):
                #name length
                f.write(struct.pack('>i', len(ktestData.objects[i][0])))
                f.write(ktestData.objects[i][0])
                #data length
                f.write(struct.pack('>i', len(ktestData.objects[i][1]))) 
                f.write(ktestData.objects[i][1])
        #print "Done ktest!"       
    #~ def _ktestToFile()

    def _parseZestiKtest(self, filename, test2zestidirMap_arg=None, \
                                                        skip_failure=False):
        """
            return a list representing the ordered list of argument 
            where each argument is represented by a pair of argtype 
            (argv or file or stdin and the corresponding sizes)
            IN KLEE, sym-files is taken into account only once (the last one)
        """

        datalist = []
        b = self.ktest_tool.KTest.fromfile(filename)
        # get the object one at the time and obtain its stuffs
        # the objects are as following: 
        #   [model_version, <the argvs> <file contain, file stat>]
        # Note that files appear in argv. 
        # here we do not have n_args because concrete (from Zesti)

        # XXX Watch this: TMP: make sure all files are at the end
        firstFile_ind = -1
        postFileArgv_ind = []
        #print(b.args) #str
        #print(b.objects) #byte
        ERROR_HANDLER.assert_true(b.objects[0][0] == b'model_version', \
                "Invalid model_version position for file: {}.\nContent:\n{}"\
                                    .format(filename, b.objects), __file__)
        #skip model_version in this loop (start from index 1)
        for pos, (name, data) in enumerate(b.objects[1:]): 
            if name == b'argv':
                if firstFile_ind >= 0:
                    postFileArgv_ind.append(pos+1)
            else:
                firstFile_ind = pos+1 if firstFile_ind < 0 else firstFile_ind
        if len(postFileArgv_ind) > 0:
            tmp_postFdat = []
            for ind_ in sorted(postFileArgv_ind, reverse=True):
                tmp_postFdat.append(b.objects[ind_])
                del b.objects[ind_]
            b.objects[firstFile_ind: firstFile_ind] = tmp_postFdat[::-1]
        #~

        # ZESTI (shadow) has issues handling forward slash(/) as argument. 
        # I think that it is a file while maybe not
        # XXX Fix that here. 
        # Since KLEE do not support directory it should be fine
        if '/' in b.args[1:]:
            for ind, (name,data) in enumerate(b.objects):
                if name == b'/':
                    if data == b'\0'*4096:
                        ERROR_HANDLER.assert_true(\
                                        b.objects[ind+1][0] == b'/-stat', \
                                        "Stat not following file", __file__)
                        del b.objects[ind:ind+2]
                    else:
                        ERROR_HANDLER.error_exit ("ERROR-BUG? data for "
                                        "forward slash not with data "
                                        "'\0'*4096 (zesti but workaround)", \
                                                                    __file__)

        seenFileStatsPos = set()
        stdin = None
        model_version_pos = -1
        fileargsposinObj_remove = []
        filesNstatsIndex = []
        maxFileSize = -1
        for ind,(name,data) in enumerate(b.objects):
            if ind in seenFileStatsPos:
                continue
            if ind == 0:
                if name != b"model_version":
                    ERROR_HANDLER.error_exit("The first argument in the "
                                    "ktest must be 'model_version'", __file__)
                else:
                    model_version_pos = ind
            else:
                # File passed , In this case, there is: 
                # (1) an ARGV obj with data the filename, 
                # (2) an Obj with name the filename and data the file data, 
                # (3) an Obj for file stat (<filename>-stat) 
                arguments = [bytes(_x, 'utf-8') for _x in b.args[1:]]
                indexes_ia = [i for i,x in enumerate(arguments) if x == name]
                
                # filename in args, the corresponding position in 
                # datalist is indexes_ia, or name is not argv but is contained.
                # example "--file=in1" TODO
                if len(indexes_ia) > 0 or name != b"argv": 
                    # in case the same name appears many times in args, 
                    # let the user manually verify
                    if len(indexes_ia) > 1:
                        if test2zestidirMap_arg is not None:
                            actual_test = None
                            for at in test2zestidirMap_arg:
                                if os.path.dirname(filename).endswith(\
                                                    test2zestidirMap_arg[at]):
                                    actual_test = at
                                    break
                            msg = " ".join(["\n>> CONFLICT: the file object",\
                                "at position ",str(ind),"with name",str(name),\
                                "in ktest",filename,"appears several times",\
                                "in args list (The actual test is:", 
                                actual_test,").\n",\
                                "    >> Please choose its space separated",\
                                "position(s), (",str(indexes_ia),"): "])
                        else:
                            msg = " ".join(["\n>> CONFLICT: the file object",\
                                "at position ",str(ind),"with name",str(name),\
                                "in ktest",filename,"appears several times",\
                                "in args list (Check",\
                                "OUTPUT/caches/test2zestidirMap.json for",\
                                "actual test).\n",\
                                "    >> Please choose its space separated",\
                                "position(s), (",str(indexes_ia),"): "])
                        raw = input(msg)
                        indinargs = [int(v) for v in raw.split()]
                        ERROR_HANDLER.assert_true(\
                                len(set(indinargs) - set(indexes_ia)) == 0, \
                                "input wrong indexes. "
                                "do not consider program name", __file__)
                    elif len(indexes_ia) == 1:
                        indinargs = indexes_ia
                    else: # name != "argv"
                        arguments = [bytes(_x, 'utf-8') for _x in b.args[1:]]
                        indexes_ia = [i for i,x in enumerate(arguments) \
                                                                if name in x]
                        if len(indexes_ia) <= 0:
                            err_msg = "Error: Must have " +\
                                    "at least one argv containing filename " +\
                                    "in its data.\n You could run with " +\
                                    "'skip_failure' enabled to neglect " +\
                                    "the error. Ktest file is "+filename
                            do_skip = False
                            if skip_failure is None:
                                do_skip = common_mix.confirm_execution(err_msg + \
                                      "\n>> DO YOU WANT TO Skip THE FAILURE? ")
                            if not do_skip and not skip_failure:
                                ERROR_HANDLER.error_exit (err_msg, __file__)
                        if len(indexes_ia) > 1:
                            if test2zestidirMap_arg is not None:
                                actual_test = None
                                for at in test2zestidirMap_arg:
                                    if os.path.dirname(filename).endswith(\
                                                    test2zestidirMap_arg[at]):
                                        actual_test = at
                                        break
                                msg = " ".join(["\n>> HINT NEEDED: the file",
                                    "object at position ",ind,"with name",\
                                    str(name),"in ktest",filename, "has file",\
                                    "with complex argv (The actual test is:",\
                                    actual_test,")."])
                            else:
                                msg = " ".join(["\n>> HINT NEEDED: the file",\
                                    "object at position ",ind,"with name",\
                                    str(name),"in ktest",filename, "has file",\
                                    "with complex argv (Check",\
                                    "OUTPUT/caches/test2zestidirMap.json",\
                                    "for actual test).\n", \
                                    "    >> Please choose its space", \
                                    "separated position(s), (",\
                                    str(indexes_ia), "):"])
                            raw = input(msg)
                            indinargs = [int(v) for v in raw.split()]
                            ERROR_HANDLER.assert_true(\
                                len(set(indinargs) - set(indexes_ia)) == 0, \
                                "input wrong indexes. do not consider "
                                "program name", __file__)
                        else:
                            indinargs = indexes_ia

                    for iv in indinargs:
                        # XXX len is 1 because this will be the file name 
                        # which is 1 char string in klee: 
                        #   'A', 'B', ... ('A' + k | 0 <= k <= 255-'A')
                        datalist[iv] = ('ARGV', 1)  

                    # ARGVs come before files in objects
                    fileargsposinObj_remove.append(indinargs)  

                    # seach for its stat, it should be the next object
                    found = False
                    if b.objects[ind + 1][0] == name + b"-stat":
                        seenFileStatsPos.add(ind + 1)
                        found = True
                    if not found:
                        ERROR_HANDLER.error_exit(\
                                "File is not having stat in ktest", __file__)

                    filesNstatsIndex += [ind, ind+1]
                    maxFileSize = max (maxFileSize, len(data))
                #elif name == "stdin-stat": #case of stdin
                #    stdin = ('STDIN', len(data)) #XXX 
                else: 
                    #ARGV
                    ERROR_HANDLER.assert_true(name == b"argv", \
                            "name ({}) not in args and not argv: {}".format(\
                                                    name, filename), __file__)
                    #XXX
                    datalist.append(('ARGV', len(data))) 

        if len(filesNstatsIndex) > 0:
            ERROR_HANDLER.assert_true(filesNstatsIndex == list(range(\
                            filesNstatsIndex[0], filesNstatsIndex[-1]+1)), \
                        "File objects are not continuous: (File {}): {}{}"\
                        .format(filename, filesNstatsIndex, list(range(\
                            filesNstatsIndex[0], filesNstatsIndex[-1]+1))),
                        __file__)

        if model_version_pos == 0:
            #for ii in range(len(fileargsposinObj_remove)):
            #    for iii in range(len(fileargsposinObj_remove[ii])):
            #        fileargsposinObj_remove[ii][iii] += 1
            # Do bothing for fileargsposinObj_remove because already 
            # indexed to not account for model version XXX
            if len(fileargsposinObj_remove) > 0:
                if len(fileargsposinObj_remove[-1]) > 0:
                    ERROR_HANDLER.assert_true(max(\
                        fileargsposinObj_remove[-1]) < filesNstatsIndex[0], \
                        "arguments do not all come before files in object", \
                        __file__)
            #-1 for model_versio which will be move to end later
            filesNstatsIndex = [(v - 1) for v in filesNstatsIndex] 

            # stdin and after files obj
            if stdin is not None:
                # -2 because of stdin and stdin-stat, -1 for mode_version
                afterLastFilenstatObj = max(filesNstatsIndex) + 1 if \
                        len(filesNstatsIndex) > 0 else (len(b.objects) - 2 -1)
                datalist.append(stdin)
            else:
                # -1 for model_version
                afterLastFilenstatObj = max(filesNstatsIndex) + 1 if \
                        len(filesNstatsIndex) > 0 else (len(b.objects) - 1) 
                # shadow-zesti have problem with stdin, \
                #use our hack on wrapper to capture that
                stdin_file = os.path.join(os.path.dirname(filename), \
                                        KTestTestFormat.STDIN_KTEST_DATA_FILE)
                ERROR_HANDLER.assert_true(os.path.isfile(stdin_file), \
                        "The stdin exported in wrapper is missing for test: "\
                        + filename, __file__)
                with open(stdin_file) as f:
                    sidat = f.read()
                    if len(sidat) > 0:
                        symin_obj = (b'stdin', bytes(sidat, 'utf-8')) #'\0'*1024)
                        syminstat_obj = (b'stdin-stat', b'\0'*144)
                        b.objects.append(symin_obj)
                        b.objects.append(syminstat_obj)
                        stdin = ('STDIN', len(sidat)) #XXX 
                        datalist.append(stdin)
        else:
            # For last, take care of putting stdin before it  
            # or last object initially"
            ERROR_HANDLER.error_exit("model version need to be either 1st", \
                                                                    __file__)

        # put 'model_version' last
        ERROR_HANDLER.assert_true(model_version_pos >= 0, \
            "'model_version' not found in ktest file: "+filename, __file__)

        b.objects.append(b.objects[model_version_pos])
        del b.objects[model_version_pos]
        
        return b, datalist, filesNstatsIndex, maxFileSize, \
                                fileargsposinObj_remove, afterLastFilenstatObj
    #~ def _parseZestiKtest()

    @staticmethod
    def _is_sym_args_having_nargs(sym_args, check_good=False):
        key_w, min_n_arg, max_n_arg, maxlen = sym_args.strip().split()
        min_n_arg, max_n_arg, maxlen = map(int, (min_n_arg, max_n_arg, maxlen))
        if check_good:
            ERROR_HANDLER.assert_true("-sym-args" in key_w, \
                            "Invalid key_w, must be having '-sym-args '", \
                            __file__)
            ERROR_HANDLER.assert_true(min_n_arg <= max_n_arg, \
                            "error: min_n_arg > max_n_arg. (bug)", __file__)
        if min_n_arg < max_n_arg:
            return True
        return False
    # def _is_sym_args_having_nargs()

    @staticmethod
    def _bestFit(outMaxVals, outNonTaken, inVals):
        assert len(inVals) <= len(outMaxVals)
        enabledArgs = [True] * len(outMaxVals)
        for i in range(len(inVals)):
            outMaxVals[i] = max(outMaxVals[i], inVals[i])
        for i in range(len(inVals), len(outMaxVals)):
            outNonTaken[i] = True
            enabledArgs[i] = False
        return enabledArgs
    #~ def _bestFit()

    def _getSymArgsFromZestiKtests (self, ktestFilesList, \
                                                    test2zestidirMap_arg, \
                                                    argv_becomes_arg_i=False, 
                                                    add_sym_stdout=False,
                                                    skip_failure=False):
        testNamesList = list(test2zestidirMap_arg.keys())
        ERROR_HANDLER.assert_true(len(ktestFilesList) == len(testNamesList), 
                "Error: size mismatch btw ktest and names: {} VS {}".format(\
                        len(ktestFilesList), len(testNamesList)), __file__)

        # XXX implement this. For program with file as parameter, 
        # make sure that the filenames are renamed in the path conditions
        # (TODO double check)
        listTestArgs = []
        ktestContains = {"CORRESP_TESTNAME":[], "KTEST-OBJ":[]}
        maxFileSize = -1
        filenstatsInObj = []
        fileArgInd = []
        afterFileNStat = []
        for ipos, ktestfile in enumerate(ktestFilesList):
            # XXX Zesti do not generate valid Ktest file when an argument 
            # is the empty string. Example tests 'basic_s18' of EXPR which 
            # is: expr "" "|" ""
            # The reson is that when writing ktest file, klee want the name 
            # to be non empty thus it fail (I think). 
            # Thus, we skip such tests here TODO: remove thes from all tests
            # so to have fair comparison with semu
            if os.system(" ".join(['ktest-tool ', ktestfile, \
                                                "> /dev/null 2>&1"])) != 0:
                logging.warning("Skipping test because Zesti generated "
                                "invalid KTEST file: " + ktestfile)
                continue
            b_tmp = self.ktest_tool.KTest.fromfile(ktestfile)
            if len(b_tmp.objects) == 0:
                logging.warning("Skipping test because Zesti "
                                "generated empty KTEST file:" + ktestfile)
                continue

            # sed because Zesti give argv, argv_1... while sym args gives 
            # arg0, arg1,...
            ktestdat, testArgs, fileNstatInd, \
                    maxFsize, fileargind, afterFnS = self._parseZestiKtest(\
                                            ktestfile, test2zestidirMap_arg, \
                                                    skip_failure=skip_failure)
            listTestArgs.append(testArgs)
            ktestContains["CORRESP_TESTNAME"].append(testNamesList[ipos])
            ktestContains["KTEST-OBJ"].append(ktestdat)
            filenstatsInObj.append(fileNstatInd)
            maxFileSize = max(maxFileSize, maxFsize)
            fileArgInd.append(fileargind)
            afterFileNStat.append(afterFnS)

        if len(listTestArgs) <= 0:
            logging.error("no ktest data, ktest PCs: " + str(ktestFilesList))
            err_msg = "No ktest data could be extracted from ktests."
            do_skip = False
            if skip_failure is None:
                do_skip = common_mix.confirm_execution(err_msg + \
                             "\n>> DO YOU WANT TO Skip THE FAILURE? ")
            if not do_skip and not skip_failure:
                ERROR_HANDLER.error_exit (err_msg, __file__)

        # update file data in objects (shortname and size)

        # divide by 2 beacause has stats
        nmax_files = int(max([len(fpv) for fpv in filenstatsInObj]) / 2) 
        if nmax_files > 0:
            shortFnames = FileShortNames().ShortNames[:nmax_files]
            for ktpos in range(len(ktestContains["CORRESP_TESTNAME"])):
                ktdat = ktestContains["KTEST-OBJ"][ktpos]

                # update file argument
                for ind_fai, fainds in enumerate(fileArgInd[ktpos]):
                    for fai in fainds:
                        # [:-1] in ktdat.objects[fai][1][-1] because of 
                        # last '\0'
                        if ktdat.objects[fai][1][:-1] == ktdat.objects\
                                    [filenstatsInObj[ktpos][2*ind_fai]][0]:
                            ktdat.objects[fai] = (ktdat.objects[fai][0], \
                                                        shortFnames[ind_fai])
                        else:
                            #print (len(ktdat.objects[fai][1]) ,\
                            #       len(ktdat.objects[filenstatsInObj[ktpos]\
                            #                               [2*ind_fai]][0])
                            #print list(ktdat.objects[fai][1]),\
                            #       list(ktdat.objects[filenstatsInObj\
                            #                        [ktpos][2*ind_fai]][0])
                            msg = " ".join(["\n>> MANUAL REPLACE: the final",\
                                "file name is "+str(shortFnames[ind_fai])+".",\
                                "initial name is "+str(ktdat.objects[\
                                filenstatsInObj[ktpos][2*ind_fai]][0])+".",
                                "\n  >> Test name is:", \
                                ktestContains["CORRESP_TESTNAME"][ktpos],\
                                "\n    >> Please replace initial file name",\
                                "with new in "+str(ktdat.objects[fai][1])+
                                " :"])
                            raw = bytes(input(msg).strip(), 'utf-8')
                            if ktdat.objects[fai][1].endswith(b'\0'):
                                raw += b'\0'
                            ktdat.objects[fai] = (ktdat.objects[fai][0], raw)

                # first add file object of additional files
                addedobj = []
                # divide by two because also has stat
                for _ in range(nmax_files - int(len(filenstatsInObj[ktpos])/2)): 
                    symf_obj = (b'', b'\0'*maxFileSize)
                    symfstat_obj = (b'-stat', b'\0'*144)
                    addedobj.append(symf_obj)
                    addedobj.append(symfstat_obj)
                insat = afterFileNStat[ktpos] 
                ktdat.objects[insat:insat] = addedobj
                filenstatsInObj[ktpos] += range(insat, insat+len(addedobj))

                # Now update the filenames and data
                for ni, fi_ in enumerate(range(0, \
                                            len(filenstatsInObj[ktpos]), 2)):
                    find_ = filenstatsInObj[ktpos][fi_]
                    fsind_ = filenstatsInObj[ktpos][fi_ + 1]
                    ERROR_HANDLER.assert_true(ktdat.objects[find_][0] \
                                    + b'-stat' == ktdat.objects[fsind_][0],
                                                        "error", __file__)
                    #file
                    ktdat.objects[find_] = (shortFnames[ni]+b"-data", \
                                        ktdat.objects[find_][1] + \
                                        b'\0'*(maxFileSize - \
                                        len(ktdat.objects[find_][1]))) 
                    #file
                    ktdat.objects[fsind_] = (shortFnames[ni]+b"-data" + \
                                        b'-stat', ktdat.objects[fsind_][1])

        # Make a general form out of listTestArgs by inserting what 
        # is needed with size 0
        # Make use of the sym-args param that can unset a param 
        # (klee care about param order)
        # Split each test args according to the FILE type (STDIN is 
        # always last), as follow: ARGV ARGV FILE ARGV FILE ...
        # then use -sym-args to flexibly set the number of enables argvs. 
        # First process the case before the first FILE, 
        # then between 1st and 2nd
        commonArgs = []
        commonArgsNumPerTest = {t: [] for t in range(len(listTestArgs))}
        testsCurFilePos = [0 for i in range(len(listTestArgs))]
        testsNumArgvs = [0 for i in range(len(listTestArgs))]
        symFileNameSize_ordered = []
        while (True):
            # Find the next non ARGV argument for all tests
            for t in range(len(testsNumArgvs)):
                nonargvfound = False
                for a in range(testsCurFilePos[t], len(listTestArgs[t])):
                    if listTestArgs[t][a][0] != "ARGV":
                        testsNumArgvs[t] = a - testsCurFilePos[t]
                        nonargvfound = True
                        break
                if not nonargvfound:
                    testsNumArgvs[t] = \
                                    len(listTestArgs[t]) - testsCurFilePos[t]
            # Rank test by num of ARGV args at this point
            indexes = list(range(len(testsNumArgvs)))
            indexes.sort(reverse=True, key=lambda x: testsNumArgvs[x])
            maxArgNum = testsNumArgvs[indexes[0]]
            maxlens = [0 for i in range(maxArgNum)]
            canDisable = [False for i in range(maxArgNum)]
            if maxArgNum > 0:
                enabledArgs = {t: None for t in range(len(testsNumArgvs))}
                for tid in indexes:
                    if testsNumArgvs[tid] == maxArgNum:
                        for pos,aid in enumerate(range(testsCurFilePos[tid], \
                                testsCurFilePos[tid] + testsNumArgvs[tid])):
                            maxlens[pos] = max(maxlens[pos], \
                                                    listTestArgs[tid][aid][1])
                        enabledArgs[tid] = [True] * maxArgNum
                    else:
                        # make the best fit on existing sizes
                        enabledArgs[tid] = self._bestFit(maxlens, canDisable, 
                                    [listTestArgs[tid][aid][1] for aid in \
                                            range(testsCurFilePos[tid], \
                                            testsCurFilePos[tid] + \
                                                        testsNumArgvs[tid])]) 
                
                # File related argument not cared about
                catchupS = len(commonArgs) - len(commonArgsNumPerTest[0])
                if catchupS > 0:
                    for t in commonArgsNumPerTest:
                        commonArgsNumPerTest[t] += [None] * catchupS

                for i in range(len(maxlens)):
                    if canDisable[i]:
                        arg = " ".join(["-sym-args 0 1", str(maxlens[i])])
                    else:
                        arg = " ".join(["-sym-arg", str(maxlens[i])])
                    # if previous is "-sym-args 0 <max-num> <size>" and 
                    # arg is also "-sym-args 0 1 <size>", with same <size>, 
                    # just update the previous
                    if len(commonArgs) > 0 \
                            and commonArgs[-1].startswith("-sym-args 0 ") \
                            and commonArgs[-1].endswith(" "+str(maxlens[i])):
                        tmpsplit = commonArgs[-1].split(' ')
                        assert len(tmpsplit) == 4
                        tmpsplit[2] = str(int(tmpsplit[2]) + 1)
                        commonArgs[-1] = " ".join(tmpsplit)
                        for t in commonArgsNumPerTest:
                            commonArgsNumPerTest[t][-1] += \
                                                        int(enabledArgs[t][i])
                    else:
                        commonArgs.append(arg)
                        for t in commonArgsNumPerTest:
                            commonArgsNumPerTest[t].append(int(\
                                                        enabledArgs[t][i]))

                # Update
                for t in range(len(testsNumArgvs)):
                    testsCurFilePos[t] += testsNumArgvs[t]

            # Process non ARGV argument stdin or file argument
            fileMaxSize =  maxFileSize
            stdinMaxSize = -1
            for t in range(len(testsNumArgvs)):
                # if the last arg was ARGV do nothing
                if testsCurFilePos[t] >= len(listTestArgs[t]):
                    continue
                # If next is FILE
                #if listTestArgs[t][testsCurFilePos[t]][0] == "FILE":
                #    fileMaxSize = max(fileMaxSize, \
                #                    listTestArgs[t][testsCurFilePos[t]][1])
                #    testsCurFilePos[t] += 1
                # If next is STDIN (last)
                #elif listTestArgs[t][testsCurFilePos[t]][0] == "STDIN":
                if listTestArgs[t][testsCurFilePos[t]][0] == "STDIN":
                    stdinMaxSize = max(stdinMaxSize, \
                                    listTestArgs[t][testsCurFilePos[t]][1])
                else:
                    ERROR_HANDLER.error_exit("unexpected arg type here: "+\
                                                "Not STDIN (type is "+\
                                listTestArgs[t][testsCurFilePos[t]][0]+")",\
                                                                    __file__)

            if fileMaxSize >= 0:
                commonArgs.append(" ".join(["-sym-files", str(nmax_files), \
                                                        str(fileMaxSize)]))
                symFileNameSize_ordered.append(fileMaxSize)
            if stdinMaxSize >= 0:
                commonArgs.append(" ".join(["-sym-stdin", str(stdinMaxSize)]))
                # Update object's stdin size. add if not present
                
                for i in range(len(ktestContains["CORRESP_TESTNAME"])):
                    siindex = len(ktestContains["KTEST-OBJ"][i].objects) - 1
                    while siindex>=0 and ktestContains["KTEST-OBJ"][i]\
                                            .objects[siindex][0] != b"stdin":
                        siindex -= 1
                    if siindex >= 0:
                        ERROR_HANDLER.assert_true(ktestContains["KTEST-OBJ"]\
                                [i].objects[siindex+1][0] == b"stdin-stat", \
                                "stdin must be followed by its stats", \
                                                                __file__)
                        pre_si_dat = \
                            ktestContains["KTEST-OBJ"][i].objects[siindex][1]
                        ktestContains["KTEST-OBJ"][i].objects[siindex] = \
                                        (b'stdin', pre_si_dat + \
                                        b"\0"*(stdinMaxSize - len(pre_si_dat)))
                    else:
                        symin_obj = (b'stdin', b'\0'*stdinMaxSize)
                        syminstat_obj = (b'stdin-stat', b'\0'*144)
                        ktestContains["KTEST-OBJ"][i].objects.insert(-1, \
                                                                    symin_obj)
                        ktestContains["KTEST-OBJ"][i].objects.insert(-1, \
                                                                syminstat_obj)
                
            break

        if add_sym_stdout:
            # Sym stdout, is this really needed
            commonArgs.append('--sym-stdout')
            # add sym-out to ktets just before the last (model_version)
            for i in range(len(ktestContains["CORRESP_TESTNAME"])):
                symout_obj = (b'stdout', b'\0'*1024)
                symoutstat_obj = (b'stdout-stat', b'\0'*144)
                ktestContains["KTEST-OBJ"][i].objects.insert(-1, symout_obj)
                ktestContains["KTEST-OBJ"][i].objects.insert(-1, \
                                                            symoutstat_obj)


        # TODO: UPDATE KTEST CONTAINS WITH NEW ARGUMENT LIST AND INSERT THE
        # "n_args" FOR '-sym-args'. ALSO PLACE 'model_version' AT THE END
        # For each Test, Change ktestContains args, go through common args 
        # and compute the different "n_args" using 'listTestArgs'  and stdin
        # stdout default.
        # Then write out the new KTEST that will be used to generate tests
        # for mutants.
        
        # XXX; The objects are ordered here in a way they the ARGV come
        # firts, then we have files, and finally model_version
        # File related argument not cared about
        catchupS = len(commonArgs) - len(commonArgsNumPerTest[0])
        if catchupS > 0:
            for t in commonArgsNumPerTest:
                commonArgsNumPerTest[t] += [None] * catchupS
        for t in commonArgsNumPerTest:
            objpos = 0
            for apos in range(len(commonArgsNumPerTest[t])):
                if commonArgs[apos].startswith("-sym-args "):
                    assert commonArgsNumPerTest[t][apos] is not None
                    if commonArgsNumPerTest[t][apos] > 0 and not \
                            ktestContains["KTEST-OBJ"][t].objects[objpos][0]\
                                                        .startswith(b"argv"):
                        logging.debug("\n\nCommonArgs: {}".format(commonArgs))
                        logging.debug("\nCommonArgsNumPerTest: {}".format(\
                                                    commonArgsNumPerTest[t]))
                        logging.debug("\nArgs: {}".format(\
                                    ktestContains["KTEST-OBJ"][t].args[1:]))
                        logging.debug("\nObjects: {}".format(\
                                    ktestContains["KTEST-OBJ"][t].objects))
                        # the name must be argv...
                        ERROR_HANDLER.error_exit(\
                                "must be argv, but found: {}".format(
                                        ktestContains["KTEST-OBJ"][t]\
                                            .objects[objpos][0]), __file__)

                    # Pad the argument data with '\0' until args maxlen
                    maxlen = int(commonArgs[apos].strip().split()[-1])
                    for sharedarg_i in range(commonArgsNumPerTest[t][apos]):
                        curlen = len(ktestContains["KTEST-OBJ"][t].objects[\
                                                    objpos + sharedarg_i][1])
                        curval = ktestContains["KTEST-OBJ"][t].objects[\
                                                        objpos + sharedarg_i]
                        #+1 Because last zero added after sym len
                        ktestContains["KTEST-OBJ"][t].objects[\
                                        objpos + sharedarg_i] = (curval[0], 
                                            curval[1]+b'\0'*(maxlen-curlen+1))

                    # Insert n_args
                    ## XXX No insertion of n_args if min_n_arg and max_n_arg 
                    # are equal
                    if self._is_sym_args_having_nargs(commonArgs[apos], \
                                                            check_good=True):
                        ktestContains["KTEST-OBJ"][t].objects.insert(\
                                    objpos, (b"n_args", struct.pack('<i', \
                                            commonArgsNumPerTest[t][apos])))       
                        objpos += 1 #pass just added 'n_args'

                    # Else no object for this, no need to advance this 
                    # (will be done bellow)
                    if commonArgsNumPerTest[t][apos] > 0: 
                        objpos += commonArgsNumPerTest[t][apos]
                # is an ARGV non file
                elif commonArgsNumPerTest[t][apos] is not None: 
                    if not ktestContains["KTEST-OBJ"][t].\
                                    objects[objpos][0].startswith(b"argv"):
                        logging.debug("\n\nCommonArgs: {}".format(commonArgs))
                        logging.debug("\nCommonArgsNumPerTest: {}".format(\
                                                    commonArgsNumPerTest[t]))
                        logging.debug("\nArgs: {}".format(\
                                    ktestContains["KTEST-OBJ"][t].args[1:]))
                        logging.debug("\nObjects: {}".format(\
                                    ktestContains["KTEST-OBJ"][t].objects))
                        # the name must be argv...
                        ERROR_HANDLER.error_exit("must be argv, but found: "\
                                    +ktestContains["KTEST-OBJ"][t].\
                                    objects[objpos][0], __file__)

                    # Pad the argument data with '\0' until args maxlen
                    maxlen = int(commonArgs[apos].strip().split(' ')[-1])
                    curlen = len(ktestContains["KTEST-OBJ"][t].\
                                                        objects[objpos][1])
                    curval = ktestContains["KTEST-OBJ"][t].objects[objpos]
                    #+1 Because last zero added after sym len
                    ktestContains["KTEST-OBJ"][t].objects[objpos] = \
                            (curval[0], curval[1]+b'\0'*(maxlen-curlen+1)) 
                    
                    objpos += 1
                else:  
                    # File or stdin, stdout
                    # TODO handle the case of files (NB: check above 
                    # how the files are recognized from zesti tests 
                    # (size may not be 0)
                    pass 

        # Change the args list in each ktest object with the common symb args 
        for ktdat in ktestContains["KTEST-OBJ"]:
            ktdat.args = ktdat.args[:1]
            for s in commonArgs:
                ktdat.args += s.strip().split(' ') 

        # Change all argv keywords into arg<i>
        if argv_becomes_arg_i:
            for ktdat in ktestContains["KTEST-OBJ"]:
                i_ = 0
                for objpos, (name, data) in enumerate(ktdat.objects):
                    if name != b"argv":
                        continue
                    ktdat.objects[objpos] = (b'arg'+bytes(str(i_), 'ascii'), \
                                                                        data)
                    i_ += 1

        return commonArgs, ktestContains
    #~ _getSymArgsFromZestiKtests()

    def _loadAndGetSymArgsFromKleeKTests(self, ktestsList, teststopdir, 
                                                must_have_model_version=True):
        '''
            the list of klee tests (ktest names) are il ktestsList. 
            Each represent the location of the ktest w.r.t. teststopdir
        '''
        # load the ktests
        commonArgs = None
        tmpsplitcommon = None
        stdin_fixed_common = None
        ktestContains = {"CORRESP_TESTNAME":[], "KTEST-OBJ":[]}
        for kt in ktestsList:
            ktestfile = os.path.join(teststopdir, kt)
            ERROR_HANDLER.assert_true(os.path.isfile(ktestfile), \
                    "Ktest file for test is missing :" + ktestfile, __file__)

            b = self.ktest_tool.KTest.fromfile(ktestfile)
            if commonArgs is None:
                commonArgs = [] 
                # make chunk (sym-args together with its params)
                anArg = None
                tmpsplitcommon = b.args[1:]
                for c in tmpsplitcommon:
                    if c.startswith('-sym') or c.startswith('--sym'):
                        if anArg is not None:
                            commonArgs.append(anArg)
                        anArg = c
                    else:
                        anArg += " " + c
                if anArg is not None:
                    commonArgs.append(anArg)
        
                # XXX fix problem with klee regarding stdin
                a_has_stdin = False
                for sa in commonArgs:
                    if '-sym-stdin ' in sa:
                        a_has_stdin = True
                        break
                if not a_has_stdin:
                    o_stdin_len = -1
                    for o in b.objects:
                        if o[0] == b'stdin':
                            o_stdin_len = len(o[1])
                            break
                    if o_stdin_len >= 0:
                        si_pos = len(commonArgs)-1 if '-sym-stdout' in \
                                        commonArgs[-1] else len(commonArgs)
                        commonArgs.insert(si_pos, \
                                            "-sym-stdin "+str(o_stdin_len))
                        stdin_fixed_common = []
                        for s_a in commonArgs:
                            stdin_fixed_common += s_a.split()
                # make sure that model_version is the last object
                ERROR_HANDLER.assert_true(not must_have_model_version or \
                                    b.objects[-1][0] == b"model_version", \
                                "The last object is not 'model_version'" + \
                                                " in klee's ktest", __file__)
            else:
                ERROR_HANDLER.assert_true(tmpsplitcommon == b.args[1:], \
                            "Sym Args are not common among KLEE's ktests", 
                                                                    __file__)
                # handle case when problem with klee and stdin
                if stdin_fixed_common is not None:
                    b.args[1:] = list(stdin_fixed_common)

            ktestContains["KTEST-OBJ"].append(b)
            ktestContains["CORRESP_TESTNAME"].append(kt)

        return commonArgs, ktestContains
    #~ def _loadAndGetSymArgsFromKleeKTests()

    def _updateObjects(self, argvinfo, ktestContains):
        '''
            Use old and new from argv_old_new to update the argv in 
            ktestContain
            Sequence in ktestobject: 
            1) argv(arg<i>), 
            2) [<ShortName>-data, <ShortName>-data-stat], 
            3) [stdin, stdin-stat], 
            4) [stdout, stdout-stat], 
            5) model_version
        '''

        def isCmdArg(name):
            if name == b'n_args' or name.startswith(b'argv') \
                                or name.startswith(b'arg'):
                return True
            return False
        #~ def isCmdArg()

        # list_new_sym_args = [(<nmin>,<nmax>,<size>),...]
        def old2new_cmdargs(objSegment, list_new_sym_args, old_has_nargs):
            res = []
            # First obj may be n_args object
            nargs = len(objSegment) - int(old_has_nargs) 
            nums = [x[0] for x in list_new_sym_args]
            n_elem = sum(nums)
            ERROR_HANDLER.assert_true(n_elem <= nargs , \
                            "min sum do not match to nargs. n_args="\
                            +str(nargs)+", min sum="+str(n_elem), __file__)
            for i in range(len(nums))[::-1]:
                rem = nargs - n_elem
                if rem <= 0:
                    break
                inc = min(rem, list_new_sym_args[i][1] - nums[i])
                n_elem += inc
                nums[i] += inc
            ERROR_HANDLER.assert_true(n_elem == nargs, \
                "n_elem must be equal to nargs here. Got: {} VS {}".format(\
                                                    n_elem, nargs), __file__)

            # put elements in res according to nums
            ao_ind = 1
            for i,n in enumerate(nums):
                if self._is_sym_args_having_nargs(" ".join(['-sym-args']\
                                    +list(map(str, list_new_sym_args[i])))):
                    res.append((b"n_args", struct.pack('<i', n)))
                for j in range(ao_ind, ao_ind+n):
                    res.append((objSegment[j][0], objSegment[j][1] \
                                         + b'\0'*(list_new_sym_args[i][2] \
                                            - len(objSegment[j][1]) + 1)))
                ao_ind += n
            assert ao_ind == len(objSegment)
            return res
        #~ def old2new_cmdargs()

        assert len(argvinfo['old']) == len(argvinfo['new'])

        argvinfo_new_extracted = []
        for vl in argvinfo['new']:
            argvinfo_new_extracted.append([])
            for v in vl:
                # assuming all args in new are sym-args (no sym-arg)
                kw, nmin, nmax, size = v.split()
                argvinfo_new_extracted[-1].append(\
                                            (int(nmin), int(nmax), int(size)))

        for okt in ktestContains['KTEST-OBJ']:
            kt_obj = okt.objects
            # First process sym-files, sym-stdin and sym-stdout
            pointer = -1 #model_version
            # sym-stdout
            if argvinfo['sym-std']['out-present']:
                if not argvinfo['sym-std']['out-present-pre']:
                    kt_obj.insert(pointer, (b'stdout', b'\0'*1024))
                    kt_obj.insert(pointer, (b'stdout-stat', b'\0'*144))
                pointer -= 2
            
            # sym-stdin
            if argvinfo['sym-std']['in-present']:
                if not argvinfo['sym-std']['in-present-pre']:
                    kt_obj.insert(pointer, \
                            (b'stdin', b'\0'*argvinfo['sym-std']['in-size']))
                    kt_obj.insert(pointer, (b'stdin-stat', b'\0'*144))
                    pointer -= 2
                elif argvinfo['sym-std']['in-size-pre'] \
                                            != argvinfo['sym-std']['in-size']:
                    # first 2 because model_version is at position -1 
                    # and secaon because of stdout-stat
                    stdin_stat_pos = pointer - 1 
                    stdin_pos = stdin_stat_pos - 1
                    ERROR_HANDLER.assert_true(\
                                        kt_obj[stdin_pos][0] == b'stdin', 
                                        "Expected stdin as object just before"
                                        "stdout and model_version", __file__)
                    kt_obj[stdin_pos] = (kt_obj[stdin_pos][0], \
                                    kt_obj[stdin_pos][1] + \
                                    b'\0'*(argvinfo['sym-std']['in-size'] - \
                                        argvinfo['sym-std']['in-size-pre']))
                    pointer -= 2
                else:
                    pointer -= 2
           
            # sym-files
            if argvinfo['sym-files']['nFiles'] > 0:
                if argvinfo['sym-files']['nFiles-pre'] > 0 and \
                                        argvinfo['sym-files']['size-pre'] \
                                            != argvinfo['sym-files']['size']:
                    for f_p in range(pointer - \
                                2*argvinfo['sym-files']['nFiles-pre'], \
                                                                pointer, 2):
                        ERROR_HANDLER.assert_true(\
                                    not isCmdArg(kt_obj[f_p][0]), \
                                    "Expected sym file object, but "+\
                                    "found cmd arg: Pos="+str(f_p)\
                                    +", object="+str(kt_obj[f_p]), __file__)
                        kt_obj[f_p] = (kt_obj[f_p][0], kt_obj[f_p][1] \
                                    + b'\0'*(argvinfo['sym-files']['size'] \
                                        - argvinfo['sym-files']['size-pre']))

                if argvinfo['sym-files']['nFiles-pre'] != \
                                            argvinfo['sym-files']['nFiles']:
                    # add empty file at position
                    snames = FileShortNames().ShortNames\
                                    [argvinfo['sym-files']['nFiles-pre']\
                                            :argvinfo['sym-files']['nFiles']]
                    for sn in snames:
                        kt_obj.insert(pointer, (sn+b'-data', \
                                        b'\0'*argvinfo['sym-files']['size']))
                        kt_obj.insert(pointer, (sn+b'-data'+b'-stat', b'\0'*144))
                pointer -= 2*argvinfo['sym-files']['nFiles']
            
            #sym cmdargv(arg)
            obj_ind = 0
            for arg_ind in range(len(argvinfo['old'])):
                #added ones, all are -sym-args, just add n_args=0
                if argvinfo['old'][arg_ind] is None: 
                    for sa in argvinfo['new'][arg_ind]:
                        ERROR_HANDLER.assert_true(\
                                        self._is_sym_args_having_nargs(sa), \
                                        "min_n_args and max_n_arg must"+\
                                        " be different here", __file__)
                        kt_obj.insert(obj_ind, \
                                            (b"n_args", struct.pack('<i', 0)))
                        obj_ind += 1
                else:
                    ERROR_HANDLER.assert_true(isCmdArg(kt_obj[obj_ind][0]), \
                        "Supposed to be CMD arg: {}".format(\
                                                kt_obj[obj_ind][0]), __file__)
                    if '-sym-arg ' in argvinfo['old'][arg_ind]:
                        ERROR_HANDLER.assert_true(\
                                    len(argvinfo['new'][arg_ind]) == 1, \
                                    "must be one sym-args here", __file__)
                        if self._is_sym_args_having_nargs(\
                                                argvinfo['new'][arg_ind][0]):
                            kt_obj.insert(obj_ind, (b"n_args", \
                                                        struct.pack('<i', 1)))
                            #Go after n_args
                            obj_ind += 1 

                        # Update the object len to the given in sym-args
                        old_len_ = len(kt_obj[obj_ind][1])
                        new_len_ = argvinfo_new_extracted[arg_ind][0][2] + 1
                        if old_len_ < new_len_:
                            kt_obj[obj_ind] = (kt_obj[obj_ind][0], \
                                                kt_obj[obj_ind][1] + \
                                                b'\0'*(new_len_ - old_len_))
                        else:
                            ERROR_HANDLER.assert_true(old_len_ == new_len_, \
                                    "new arg len lower than pld len (BUG)", \
                                                                    __file__)
                        #Go after argv(arg)
                        obj_ind += 1 
                    else: 
                        #sym-args
                        if self._is_sym_args_having_nargs(\
                                                argvinfo['old'][arg_ind]):
                            ERROR_HANDLER.assert_true(\
                                        kt_obj[obj_ind][0] == b'n_args', \
                                            "must be n_args here", __file__)
                            nargs = struct.unpack('<i', kt_obj[obj_ind][1])[0]
                            old_has_nargs = True
                        else:
                            #n_args == min_n_arg == max_n_arg
                            nargs = int(argvinfo['old'][arg_ind]\
                                                        .strip().split()[1]) 
                            old_has_nargs = False

                        if len(argvinfo['new'][arg_ind]) > 1 \
                                        or argvinfo['old'][arg_ind] \
                                            != argvinfo['new'][arg_ind][0]:
                            # put the args enabled as match to 
                            # the right most in new
                            tmppos_last = obj_ind + nargs 
                            replacement = old2new_cmdargs(\
                                            kt_obj[obj_ind:tmppos_last+1], \
                                            argvinfo_new_extracted[arg_ind], \
                                            old_has_nargs=old_has_nargs)
                            kt_obj[obj_ind:tmppos_last+1] = replacement
                            obj_ind += len(replacement)
                        else:
                            #+1 for n_args obj
                            obj_ind += nargs + int(old_has_nargs) 

            ERROR_HANDLER.assert_true(len(kt_obj) + pointer == obj_ind, \
                                "Some argv objects were not considered? ("\
                                    +str(len(kt_obj))+", "+str(pointer)+", "\
                                                +str(obj_ind)+")", __file__)
    #~ def _updateObjects()

    def _mergeZestiAndKleeKTests (self, outDir, ktestContains_zest, \
                        commonArgs_zest, ktestContains_klee, commonArgs_klee):
        commonArgs = []
        name2ktestMap = {}
        ktestContains = {"CORRESP_TESTNAME":[], "KTEST-OBJ":[]}

        if ktestContains_zest is None:
            ERROR_HANDLER.assert_true(commonArgs_zest is None, "", __file__)
            ERROR_HANDLER.assert_true(ktestContains_klee is not None \
                                and commonArgs_klee is not None, "", __file__)
            ktestContains = ktestContains_klee
            commonArgs = commonArgs_klee
        elif ktestContains_klee is None:
            ERROR_HANDLER.assert_true(commonArgs_klee is None, "", __file__)
            ERROR_HANDLER.assert_true(ktestContains_zest is not None \
                                and commonArgs_zest is not None, "", __file__)
            ktestContains = ktestContains_zest
            commonArgs = commonArgs_zest
        else:
            def getSymArgvParams(symargstr):
                outdict = {}
                tz = symargstr.split()
                if len(tz) == 2: #sym-arg
                    outdict['min'] = outdict['max'] = 1
                    outdict['size'] = int(tz[-1])
                else: #sym-args
                    outdict['min'] = int(tz[-3])
                    outdict['max'] = int(tz[-2])
                    outdict['size'] = int(tz[-1])
                return outdict
            #~ def getSymArgvParams()


            assert commonArgs_zest is not None and commonArgs_klee is not None
            # Common Args must have either sym-arg or sym-args 
            # or sym-files or sym-stdin or sym-stdout
            for common in [commonArgs_zest, commonArgs_klee]:
                for a in common:
                    if "-sym-arg " in a or "-sym-args " in a \
                            or "-sym-files " in a or "-sym-stdin " in a \
                                                        or "-sym-stdout" in a:
                        continue
                    ERROR_HANDLER.error_exit ("Unsupported symbolic "+\
                                "argument: "+a+", in "+str(common), __file__)

            # create merged commonArgs and a map between both zest 
            # and klee commonargs to new commonargs
            argv_zest = {
                    'old':[x for x in commonArgs_zest if "-sym-arg" in x], 
                    'new':[]}
            argv_zest['new'] = [[] for x in argv_zest['old']]
            argv_klee = {
                    'old':[x for x in commonArgs_klee if "-sym-arg" in x], 
                    'new':[]}
            argv_klee['new'] = [[] for x in argv_klee['old']]
            z_ind = 0
            k_ind = 0
            z_cur_inf = None
            k_cur_inf = None
            while True:
                if z_ind < len(argv_zest['old']) \
                                            and k_ind < len(argv_klee['old']):
                    if z_cur_inf is None:
                        z_cur_inf = getSymArgvParams (argv_zest['old'][z_ind])
                    if k_cur_inf is None:
                        k_cur_inf = getSymArgvParams (argv_klee['old'][k_ind])
                    m_min = min(z_cur_inf['min'], k_cur_inf['min'])
                    m_max = min(z_cur_inf['max'], k_cur_inf['max'])
                    M_size = max(z_cur_inf['size'], k_cur_inf['size'])
                    newarg = " ".join(["-sym-args", str(m_min), \
                                                    str(m_max), str(M_size)])

                    # add the new args to each as new, add also to commonArgs
                    commonArgs.append(newarg)
                    argv_zest['new'][z_ind].append(newarg)
                    argv_klee['new'][k_ind].append(newarg)

                    # Update the cur_infs and index
                    if z_cur_inf['max'] == m_max:
                        z_cur_inf = None
                        z_ind += 1
                    else:
                        z_cur_inf['min'] = max(0, z_cur_inf['min'] - m_max)
                        z_cur_inf['max'] = z_cur_inf['max'] - m_max
                    if k_cur_inf['max'] == m_max:
                        k_cur_inf = None
                        k_ind += 1
                    else:
                        k_cur_inf['min'] = max(0, k_cur_inf['min'] - m_max)
                        k_cur_inf['max'] = k_cur_inf['max'] - m_max

                else:
                    # handle inequalities
                    if z_ind < len(argv_zest['old']):
                        argv_klee['old'].append(None)
                        argv_klee['new'].append([])
                        if z_cur_inf is not None:
                            ERROR_HANDLER.assert_true(type(z_cur_inf) == dict,\
                                                "Invalid z_cur_inf", __file__)
                            newarg = " ".join(["-sym-args", \
                                                    str(z_cur_inf['min']), \
                                                    str(z_cur_inf['max']), \
                                                    str(z_cur_inf['size'])])
                            commonArgs.append(newarg)
                            argv_zest['new'][z_ind].append(newarg)
                            argv_klee['new'][k_ind].append(newarg)
                            z_ind += 1
                        #replace sym-arg with sym-args 0 1 because 
                        # not present in other
                        ineqs = [a.replace('sym-arg ', 'sym-args 0 1 ') \
                                        for a in argv_zest['old'][z_ind:]] 
                        commonArgs += ineqs
                        argv_zest['new'][z_ind:] = [[v] for v in ineqs]
                        argv_klee['new'][k_ind] += ineqs
                        break
                    if k_ind < len(argv_klee['old']):
                        argv_zest['old'].append(None)
                        argv_zest['new'].append([])
                        if k_cur_inf is not None:
                            ERROR_HANDLER.assert_true(type(z_cur_inf) == dict,\
                                                "Invalid z_cur_inf", __file__)
                            newarg = " ".join(["-sym-args", \
                                                    str(k_cur_inf['min']), \
                                                    str(k_cur_inf['max']), \
                                                    str(k_cur_inf['size'])])
                            commonArgs.append(newarg)
                            argv_klee['new'][k_ind].append(newarg)
                            argv_zest['new'][z_ind].append(newarg)
                            k_ind += 1
                        # replace sym-arg with sym-args 0 1 because 
                        # not present in other
                        ineqs = [a.replace('sym-arg ', 'sym-args 0 1 ') \
                                        for a in argv_klee['old'][k_ind:]] 
                        commonArgs += ineqs
                        argv_klee['new'][k_ind:] = [[v] for v in ineqs]
                        argv_zest['new'][z_ind] += ineqs
                        break
                    break

            # get sym-stdin, sym-stdout and symfiles
            argv_zest["sym-files"] = {'nFiles-pre':0, 'nFiles':0, 
                                                    'size-pre':0, 'size':0}
            argv_klee["sym-files"] = {'nFiles-pre':0, 'nFiles':0, 
                                                    'size-pre':0, 'size':0}
            argv_zest["sym-std"] = {'in-present-pre': False, 
                                    'in-size-pre':0, 'out-present-pre':False,
                                    'in-present': False, 'in-size':0, 
                                    'out-present':False}
            argv_klee["sym-std"] = {'in-present-pre': False, 'in-size-pre':0,
                                    'out-present-pre':False, 
                                    'in-present': False, 'in-size':0, 
                                    'out-present':False}
            for common, s_argv in [(commonArgs_zest, argv_zest), \
                                            (commonArgs_klee, argv_klee)]:
                # klee considers the last sym-files and stdin so we 
                # just check from begining to end
                for a in common:
                    if '-sym-files ' in a:
                        _, n_f, f_s = a.split()
                        s_argv["sym-files"]["nFiles-pre"] = int(n_f)
                        s_argv["sym-files"]["size-pre"] = int(f_s)
                    elif '-sym-stdin ' in a:
                        s_argv['sym-std']['in-present-pre'] = True
                        s_argv['sym-std']['in-size-pre'] = int(a.split()[-1])
                    elif '-sym-stdout' in a:
                        s_argv['sym-std']['out-present-pre'] = True
            argv_zest['sym-files']['nFiles'] = \
                        argv_klee['sym-files']['nFiles'] = \
                                max(argv_zest['sym-files']['nFiles-pre'], \
                                        argv_klee['sym-files']['nFiles-pre'])
            argv_zest['sym-files']['size'] = \
                            argv_klee['sym-files']['size'] = \
                                max(argv_zest['sym-files']['size-pre'], 
                                        argv_klee['sym-files']['size-pre'])
            argv_zest['sym-std']['in-size'] = \
                            argv_klee['sym-std']['in-size'] = \
                                max(argv_zest['sym-std']['in-size-pre'], \
                                        argv_klee['sym-std']['in-size-pre'])
            argv_zest['sym-std']['in-present'] = \
                        argv_klee['sym-std']['in-present'] = \
                                (argv_zest['sym-std']['in-present-pre'] \
                                    or argv_klee['sym-std']['in-present-pre'])
            argv_zest['sym-std']['out-present'] = \
                        argv_klee['sym-std']['out-present'] = \
                                (argv_zest['sym-std']['out-present-pre'] or \
                                    argv_klee['sym-std']['out-present-pre'])

            if argv_zest['sym-files']['nFiles'] > 0:
                commonArgs.append(" ".join(['-sym-files', \
                            str(argv_zest['sym-files']['nFiles']), \
                                    str(argv_zest['sym-files']['size'])]))
            if argv_zest['sym-std']['in-present']:
                commonArgs.append(" ".join(['-sym-stdin', \
                                    str(argv_zest['sym-std']['in-size'])]))
            if argv_zest['sym-std']['out-present']:
                commonArgs.append("-sym-stdout")

            self._updateObjects (argv_zest, ktestContains_zest)
            self._updateObjects (argv_klee, ktestContains_klee)

            # Merge the two objects
            ktestContains["CORRESP_TESTNAME"] = list(\
                            ktestContains_zest["CORRESP_TESTNAME"] + \
                                    ktestContains_klee["CORRESP_TESTNAME"])
            ktestContains["KTEST-OBJ"] = list(\
                                    ktestContains_zest["KTEST-OBJ"] + \
                                            ktestContains_klee["KTEST-OBJ"])

            # Change the args list in each ktest object with the 
            # common symb args 
            for ktdat in ktestContains["KTEST-OBJ"]:
                ktdat.args = ktdat.args[:1]
                for s in commonArgs:
                    ktdat.args += s.strip().split(' ') 

        # Write the new ktest files
        for i in range(len(ktestContains["CORRESP_TESTNAME"])):
            outFname = os.path.join(outDir, "test"+str(i)+".ktest")
            self._ktestToFile(ktestContains["KTEST-OBJ"][i], outFname)
            name2ktestMap[ktestContains["CORRESP_TESTNAME"][i]] = outFname

        return commonArgs, name2ktestMap
    #~ def _mergeZestiAndKleeKTests ()
#~ class ConvertCollectKtestsSeeds



import muteria.common.mix as common_mix
ERROR_HANDLER = common_mix.ErrorHandler

class DriverConfigKlee:
    def __init__(self, no_duptest_with_seed=True, verbose_generation=False,\
                                      suppress_generation_stdout=False, **kwargs):
        ERROR_HANDLER.assert_true(type(no_duptest_with_seed) == bool, \
                "invalid no_duptest_with_seed type. Must be bool", __file__)
        ERROR_HANDLER.assert_true(type(verbose_generation) == bool, \
                "invalid verbose_generation type. Must be bool", __file__)
        ERROR_HANDLER.assert_true(type(suppress_generation_stdout) == bool, \
                "invalid suppress_generation_stdout type. Must be bool", __file__)
        self.no_duptest_with_seed = no_duptest_with_seed
        self.verbose_generation = verbose_generation
        self.suppress_generation_stdout = suppress_generation_stdout
    #~ def __init__()

    def get_gen_tests_no_dup_with_seeds (self):
        return self.no_duptest_with_seed
    #~ def get_gen_tests_no_dup_with_seeds()
    
    def get_verbose_generation (self):
        return self.verbose_generation
    #~ def get_verbose_generation()
    
    def get_suppress_generation_stdout (self):
        return self.suppress_generation_stdout
    #~ def get_suppress_generation_stdout()
#~ class DriverConfigKlee



import muteria.common.mix as common_mix
ERROR_HANDLER = common_mix.ErrorHandler

class DriverConfigGCov:
    def __init__(self, allow_missing_coverage=False, use_gdb_wrapper=True, \
                                                                    **kwargs):
        ERROR_HANDLER.assert_true(type(allow_missing_coverage) == bool, \
                "invalid allow_missing_coverage type. Must be bool", __file__)
        self.allow_missing_coverage = allow_missing_coverage
        self.use_gdb_wrapper = use_gdb_wrapper
    #~ def __init__()

    def get_allow_missing_coverage(self):
        return self.allow_missing_coverage
    #~ def get_allow_missing_coverage()
    
    def get_use_gdb_wrapper(self):
        return self.use_gdb_wrapper
    #~ def get_use_gdb_wrapper()
#~ class DriverConfigGCov




from __future__ import print_function

import os
import sys
import re
import shutil
import shlex
import logging
import subprocess

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.repositoryandcode.callback_object import DefaultCallbackObject

from muteria.drivers.criteria.base_testcriteriatool import BaseCriteriaTool
from muteria.drivers.criteria import TestCriteria
from muteria.drivers import DriversUtils
from muteria.drivers.criteria.criteria_info import MutantsInfoObject

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaToolGPMutation(BaseCriteriaTool):
    def __init__(self, *args, **kwargs):
        BaseCriteriaTool.__init__(self, *args, **kwargs)
        self.instrumentation_details = os.path.join(\
                    self.instrumented_code_storage_dir, '.instru.meta.json')
        #self.wm_res_log_file = \
        #            "label_log."+TestCriteria.WEAK_MUTATION.get_field_value()
        #self.mcov_res_log_file = \
        #            "label_log."+TestCriteria.MUTANT_COVERAGE.get_field_value()
        self.mutant_data = os.path.join(self.instrumented_code_storage_dir,\
                                                            "mutant_data")
        self.gpmutation_out = os.path.join(self.mutant_data, 'gpmutation-out')
        self.separate_muts_folder_name = 'mutants.out'
        self.separate_muts_dir = os.path.join(self.gpmutation_out, \
                                                self.separate_muts_folder_name)
        self.archive_separated = True
    #~ def __init__()

    def _get_default_params(self):
        bool_params = {
            #'-keep-mutants-bc': None,
            #'-no-COV': None,
            #'-no-Meta': None,
            #'-no-WM': None, 
            #'-no-mutant-info': None,
            #'-print-preTCE-Meta': None,
            #'-write-mutants': None,
        }
        key_val_params = {
            #'-linking-flags': None,
            #'-mutant-config': None,
            #'-mutant-scope': None,
        }
        return bool_params, key_val_params
    #~ def _get_default_params()

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in ('gpmutation',):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[0]):
                return False
        return True
    #~ def installed()

    @classmethod
    def _get_meta_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in same file
            :return: list of citeria
        """
        return [
                #TestCriteria.MUTANT_COVERAGE,
                #TestCriteria.WEAK_MUTATION,
               ]
    #~ def _get_meta_instrumentation_criteria()

    @classmethod
    def _get_separated_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in different files
            :return: list of citeria
        """
        return [TestCriteria.STRONG_MUTATION]
    #~ def _get_separated_instrumentation_criteria()

    def get_instrumented_executable_paths_map(self, enabled_criteria):
        ERROR_HANDLER.error_exit(
                        "TODO: Not supported yet (single instrumented)",
                        __file__)

        '''crit_to_exes_map = {}
        obj = common_fs.loadJSON(self.instrumentation_details)
        #exes = [p for _, p in list(obj.items())]
        for c, c_exes in list(obj.items()):
            for k in c_exes:
                c_exes[k] = os.path.join(self.gpmutation_out, c_exes[k])

        for criterion in enabled_criteria:
            ERROR_HANDLER.assert_true(criterion.get_str() in obj, 
                            'criterion was not enabled during instrumentation'
                            ' with GPMutation: {}. Eneble it and run again'.format(\
                                                criterion.get_str()), __file__)
            crit_to_exes_map[criterion] = obj[criterion.get_str()]
        return crit_to_exes_map'''
    #~ def get_instrumented_executable_paths_map()

    def get_criterion_info_object(self, criterion):
        try:
            return self.mutant_info_object
        except AttributeError:
            minf_obj = MutantsInfoObject()

            gpmutation_inf_obj = common_fs.loadJSON(os.path.join(
                                    self.gpmutation_out, "mutantsInfo.json"))

            # Add elements
            for mid, info in list(gpmutation_inf_obj.items()):
                minf_obj.add_element(mid, mutant_type=info['Type'], \
                                        mutant_locs=info['SrcLoc'], \
                                        mutant_function_name=info['FuncName'])
            self.mutant_info_object = minf_obj
            return minf_obj
    #~ def get_criterion_info_object()

    def _get_single_exe_filename(self, criterion):
        ERROR_HANDLER.error_exit(
                        "TODO: Not supported yet (single instrumented)",
                        __file__)
        '''try:
            return self.single_exe_filename
        except AttributeError:
            inst_path_map = \
                        self.get_instrumented_executable_paths_map([criterion])
            tmp = inst_path_map[criterion]
            r_file = list(tmp.keys())[0]
            filename = os.path.splitext(os.path.basename(tmp[r_file]))[0]
            self.single_exe_filename = {r_file: filename}
            return self.single_exe_filename.copy()'''
    #~ def _get_single_exe_filename()

    def _get_criterion_element_executable_path(self, criterion, element_id):
#        self.sm_separate_exes = 
        ERROR_HANDLER.assert_true(self.get_criterion_info_object(criterion).\
                                            has_element(element_id),\
                        "Inexistant mutant id: "+element_id, __file__)
        
        rel_names = []
        mut_code = {}
        with open(self.instrumentation_details) as f:
            map_keys = f.readlines()
            ERROR_HANDLER.assert_true(len(map_keys) == 1, \
                            "only one executable supported for now", __file__)
            map_key = map_keys[0].strip()
        mut_code[map_key] = os.path.join(self.separate_muts_dir, element_id, \
                                                    os.path.basename(map_key)) 
        rel_names.append(os.path.join(self.separate_muts_folder_name, \
                                        element_id, os.path.basename(map_key)))
        # If archiving
        if self.archive_separated:
            archive_path = common_fs.TarGz.get_archive_filename_of(\
                                                        self.separate_muts_dir)
            ERROR_HANDLER.assert_true(os.path.isfile(archive_path), \
                                    "Archived separated mutant file missing",\
                                    __file__)
            if os.path.isdir(self.separate_muts_dir):
                shutil.rmtree(self.separate_muts_dir)
            # Extract the selected
            for arch_name in rel_names:
                err_msg = common_fs.TarGz.extractFromArchive(archive_path, \
                                                                    arch_name)
                ERROR_HANDLER.assert_true(err_msg is None, \
                            "failed to extract, err: "+str(err_msg), __file__)
        return mut_code
    #~ def _get_criterion_element_executable_path()

    def _get_criterion_element_environment_vars(self, criterion, element_id):
        '''
            return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        return None
    #~ def _get_criterion_element_environment_vars()

    def _get_criteria_environment_vars(self, result_dir_tmp, enabled_criteria):
        '''
        return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        res = {}
        for c in enabled_criteria:
            if c == TestCriteria.WEAK_MUTATION:
                ERROR_HANDLER.error_exit(
                        "TODO: Not supported yet (single instrumented)",
                        __file__)
            elif c == TestCriteria.MUTANT_COVERAGE:
                ERROR_HANDLER.error_exit(
                        "TODO: Not supported yet (single instrumented)",
                        __file__)
            else:
                res[c] = None
        return res
    #~ def _get_criteria_environment_vars()

    def _collect_temporary_coverage_data(self, criteria_name_list, \
                                            test_execution_verdict, \
                                            used_environment_vars, \
                                            result_dir_tmp, \
                                            testcase):
        ''' Get the list of mutants id weakly killed or covered
        '''
        pass
    #~ def _collect_temporary_coverage_data()

    def _extract_coverage_data_of_a_test(self, enabled_criteria, \
                                    test_execution_verdict, result_dir_tmp):
        ''' read json files and extract data
            return: the dict of criteria with covering count
        '''
        ERROR_HANDLER.assert_true(TestCriteria.STRONG_MUTATION not in \
                                                        enabled_criteria, \
                            "SM metamutant run not yet supported", __file__)

        ERROR_HANDLER.error_exit(
                        "TODO: Not supported yet (single instrumented)",
                        __file__)
        '''
        def extract_covered(filename, criterion):
            mutant_id_set = set(self.get_criterion_info_object(criterion).\
                                                        get_elements_list())
            cov_res = {
                    m: common_mix.GlobalConstants.ELEMENT_NOTCOVERED_VERDICT\
                                                    for m in mutant_id_set}
            if os.path.isfile(filename):
                with open(filename) as f:
                    for line in f:
                        mut_id = line.strip()
                        # use if because gpmutation currently do not update 
                        # WM and MCOV after fdupes TCE
                        if mut_id in mutant_id_set: 
                            cov_res[mut_id] = common_mix.GlobalConstants\
                                                    .ELEMENT_COVERED_VERDICT
            return cov_res
        #~ def extract_covered()

        res = {}
        if TestCriteria.WEAK_MUTATION in enabled_criteria:
            res[TestCriteria.WEAK_MUTATION] = extract_covered(\
                        os.path.join(result_dir_tmp, self.wm_res_log_file), \
                                                    TestCriteria.WEAK_MUTATION)
        if TestCriteria.MUTANT_COVERAGE in enabled_criteria:
            res[TestCriteria.MUTANT_COVERAGE] = extract_covered(\
                        os.path.join(result_dir_tmp, self.mcov_res_log_file), \
                                                TestCriteria.MUTANT_COVERAGE)

        return res
        '''
    #~ def _extract_coverage_data_of_a_test()

    def _do_instrument_code (self, exe_path_map, \
                                        code_builds_factory, \
                                        enabled_criteria, parallel_count=1):
        # Setup
        if os.path.isdir(self.instrumented_code_storage_dir):
            shutil.rmtree(self.instrumented_code_storage_dir)
        os.mkdir(self.instrumented_code_storage_dir)
        if os.path.isdir(self.mutant_data):
            shutil.rmtree(self.mutant_data)
        os.mkdir(self.mutant_data)

        prog = 'gpmutation'
        if self.custom_binary_dir is not None:
            prog = os.path.join(self.custom_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                                        .format(os.path.basename(prog), \
                                            self.custom_binary_dir), __file__)

        exes, _ = code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()

        ERROR_HANDLER.assert_true(len(exes) == 1, \
                                        "Support only a singe exe", __file__)

        gpmutation_subj = os.path.basename(exes[0])
        args=[gpmutation_subj, self.separate_muts_dir]

        # Execute GPMutation
        ret, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                                        prog, args_list=args)
        if (ret != 0):
            logging.error(out)
            logging.error(err)
            logging.error("\n>> CMD: " + " ".join([prog]+args) + '\n')
            ERROR_HANDLER.error_exit("gpmutation failed!", __file__)
        
        # write down the rel_path_map
        ERROR_HANDLER.assert_true(not os.path.isfile(\
                self.instrumentation_details), "must not exist here", __file__)
        
        with open (self.instrumentation_details, 'w') as f:
            for exe in exes:
                f.write(exe+'\n')

        # Archive separated if on
        if self.archive_separated:
            err_msg = common_fs.TarGz.compressDir(self.separate_muts_dir, \
                                                    remove_in_directory=True)
            ERROR_HANDLER.assert_true(err_msg is None,\
                                "Compression failed: "+str(err_msg), __file__)
    #~ def _do_instrument_code()

    ## Extra functions for gpmutation
    def get_test_gen_metamutant_bc(self):
        ERROR_HANDLER.error_exit(
                        "TODO: Not supported yet (single instrumented)",
                        __file__)
        '''
        ERROR_HANDLER.assert_true(os.path.isfile(self.instrumentation_details), \
		 "Are you sure mutant generation with GPMutation was ran?"
                 "This is needed before meta mu bc is called", __file__)
        crit2file = self.get_instrumented_executable_paths_map(\
                                            (TestCriteria.STRONG_MUTATION,))
        sm_map = list(crit2file[TestCriteria.STRONG_MUTATION].items())
        ERROR_HANDLER.assert_true(len(sm_map) == 1, \
                                "Expects exactly 1 file for Strong Mutation")
        exe_name, filepath = sm_map[0]
        return filepath + '.bc'
        '''
    #~ def get_test_gen_metamutant_bc()
#~ class CriteriaToolGPMutation



""" Typical approach to bypass doing 'make check TESTS=...' 
    for each individual test.
    
    Get the env vars set by make check TESTS and just pass them when 
    running a test.

    make the test runner call the function 'get_make_check_tests_env_vars'
    in your dev test runner and use it when running the test with
    bash or perl.
"""

import os
import json


make_check_test_env_keys = [
                "VERSION", "LOCALE_FR", "LOCALE_FR_UTF8", "abs_top_builddir", \
                "abs_top_srcdir", "abs_srcdir", "built_programs", "host_os", \
                "host_triplet", "srcdir", "top_srcdir", "CONFIG_HEADER", \
                #"CU_TEST_NAME", \
                "CC", "AWK", "EGREP", "EXEEXT", "MAKE", "PACKAGE_VERSION", \
                "PERL", "PREFERABLY_POSIX_SHELL", \
                #"PATH", \ #XXX Should we add PATH?
                "PROG"] 
make_check_test_env_vars = None


def compute_make_check_tests_env_vars(repo_make_check_test_dir, \
                                                different_make_build_dir=None):
    """
    """
    global make_check_test_env_vars
    # obtain it
    tmp_getenv_hook = os.path.join(repo_make_check_test_dir, \
                                                'my_conf_test_hook.conf.tmp')
    tmp_getenv_hook_sh = tmp_getenv_hook + '.sh'
    tmp_getenv_hook_json = tmp_getenv_hook + '.json'
    with open(tmp_getenv_hook_sh, 'w') as f:
        contain_str = "#! /bin/bash\n\n"
        contain_str += "python -c \"import os; import json; " \
                + "f = open('{}', 'w'); ".format(tmp_getenv_hook_json) \
                + "json.dump(dict(os.environ), f); " \
                + "f.close()\"\n"
        contain_str += "exit $?"
        f.write(contain_str+'\n')
    os.system('chmod +x '+tmp_getenv_hook_sh)
    cwd = os.getcwd()
    os.chdir(repo_make_check_test_dir)
    test_cmd = 'make check TESTS='+tmp_getenv_hook_sh
    os.system(test_cmd)
    if not os.path.isfile(tmp_getenv_hook_json):
        if different_make_build_dir:
            os.chdir(different_make_build_dir)
        os.system('make')
        #if different_make_build_dir:
        #    os.chdir(repo_make_check_test_dir)
        os.system(test_cmd)
    os.chdir(cwd)
    assert os.path.isfile(tmp_getenv_hook_json), \
                                        "Getting makefile runtest env failed."\
                            " File {} is missing".format(tmp_getenv_hook_json)
    with open(tmp_getenv_hook_json) as f: 
        make_check_test_env_vars = json.load(f)
    useless_make_test_env_keys = set(make_check_test_env_vars) \
                                    - set(make_check_test_env_keys)
    for k in useless_make_test_env_keys:
        del make_check_test_env_vars[k]
    # TODO: failure error handling
    
    # cleanup
    os.remove(tmp_getenv_hook_sh)
    os.remove(tmp_getenv_hook_json)
    if os.path.isfile(tmp_getenv_hook+'.log'):
        os.remove(tmp_getenv_hook+'.log')
    if os.path.isfile(tmp_getenv_hook+'.trs'):
        os.remove(tmp_getenv_hook+'.trs')
#~ def compute_make_check_tests_env_vars()

def get_make_check_tests_env_vars():
    return make_check_test_env_vars
#~ def get_make_check_tests_env_vars()




import os
import muteria.common.mix as common_mix
ERROR_HANDLER = common_mix.ErrorHandler

from muteria.drivers.testgeneration.tools_by_languages.c.klee.driver_config \
                                                        import DriverConfigKlee

class MetaMuSource(common_mix.EnumAutoName):
    MART = 'mart'
    ANNOTATION = 'manual_annotation'
#~ class MetaMuSource

class DriverConfigSemu(DriverConfigKlee):
    def __init__(self, max_mutant_count_per_cluster=100,
                        meta_mutant_source=MetaMuSource.MART,
                        target_only_live_mutants=True, **kwargs):
        DriverConfigKlee.__init__(self, **kwargs)
        ERROR_HANDLER.assert_true(max_mutant_count_per_cluster > 0, \
                        "max_mutant_count_per_cluster must be > 0", __file__)
        if type(meta_mutant_source) == str \
                                    and meta_mutant_source.endswith(".bc"):
            # Consider case when directly specifying meta-mu file in semu
            ERROR_HANDLER.assert_true(os.path.isfile(meta_mutant_source), 
                    "Specified meta-mu bc file is missing: "+meta_mutant_source)
            meta_mutant_source = os.path.realpath(meta_mutant_source)
        else:
            ERROR_HANDLER.assert_true(
                MetaMuSource.is_valid(meta_mutant_source), \
                        "invalid Meta mu source. must be instance of "
                        "MetaMuSource enum", 
                __file__
            )
        self.max_mutant_count_per_cluster = max_mutant_count_per_cluster
        self.meta_mutant_source = meta_mutant_source
        self.target_only_live_mutants = target_only_live_mutants 
    #~ def __init__()

    def requires_criteria_instrumented(self):
        if type(self.meta_mutant_source) == str or meta_mutant_source == MetaMuSource.ANNOTATION:
            return False
        return True
    #~ def requires_criteria_instrumented()
    
    def get_max_mutant_count_per_cluster (self):
        return self.max_mutant_count_per_cluster
    #~ def get_max_mutant_count_per_cluster ()

    def get_meta_mutant_source (self):
        return self.meta_mutant_source
    #~ def get_meta_mutant_source ()

    def get_target_only_live_mutants(self):
        return self.target_only_live_mutants
    #~ def get_target_only_live_mutants()
#~ class DriverConfigSemu



""" TODO: call coverage
"""

from __future__ import print_function

import os
import sys
import re
import shutil
import shlex
import logging
import subprocess

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.repositoryandcode.callback_object import DefaultCallbackObject

from muteria.drivers.criteria.base_testcriteriatool import BaseCriteriaTool
from muteria.drivers.criteria import TestCriteria
from muteria.drivers import DriversUtils

from muteria.drivers.criteria.tools_by_languages.c.gcov.driver_config \
                                                        import DriverConfigGCov


ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaToolGCov(BaseCriteriaTool):
    def __init__(self, *args, **kwargs):
        BaseCriteriaTool.__init__(self, *args, **kwargs)

        self.driver_config = None
        if self.config.get_tool_user_custom() is not None:
            self.driver_config = \
                            self.config.get_tool_user_custom().DRIVER_CONFIG
        if self.driver_config is None:
            self.driver_config = DriverConfigGCov()
        else:
            ERROR_HANDLER.assert_true(isinstance(self.driver_config, \
                                        DriverConfigGCov),\
                                            "invalid driver config", __file__)

        self.instrumentation_details = os.path.join(\
                    self.instrumented_code_storage_dir, '.instru.meta.json')
        self.gcov_files_list_filename = "gcov_files.json"
        self.gc_files_dir = os.path.join(\
                                        self.criteria_working_dir, "gcno_gcda")
        self.gcov_gdb_wrapper_template = os.path.join(\
                                                os.path.dirname(__file__), \
                                                'gcov-gdb_call_wrapper.sh.in')
        self.gcov_gdb_wrapper_sh = os.path.join(\
                                        self.instrumented_code_storage_dir, \
                                                    "tmp_gcov_gdb_wrapper.sh")

        # clean any possible gcda file
        for file_ in self._get_gcda_list():
            os.remove(file_)
        for file_ in self._get_gcov_list():
            os.remove(file_)
    #~ def __init__()

    def _get_gcov_list(self):
        gcov_files = self._recursive_list_files(self.gc_files_dir, '.gcov')
        return gcov_files
    #~ def _get_gcov_list()

    def _get_gcda_list(self):
        gcda_files = self._recursive_list_files(self.gc_files_dir, '.gcda')
        return gcda_files
    #~ def _get_gcda_list()

    @staticmethod
    def _recursive_list_files(topdir, file_suffix):
        filtered_files = []
        for root, dirs, files in os.walk(topdir):
            for file_ in files:
                if file_.endswith(file_suffix):
                    filtered_files.append(os.path.join(root, file_))
        return filtered_files
    #~ def _recursive_list_files()

    class InstrumentCallbackObject(DefaultCallbackObject):
        def after_command(self):
            if self.op_retval != common_mix.GlobalConstants.COMMAND_SUCCESS:
                ERROR_HANDLER.error_exit("Build failed", __file__)
            gc_files_dir, rel_path_map = self.post_callback_args
            for _, obj in list(self.source_files_to_objects.items()):
                rel_raw_filename, _ = os.path.splitext(obj)
                gcno_file = rel_raw_filename + '.gcno' 
                relloc = os.path.join(gc_files_dir, os.path.dirname(obj))
                if not os.path.isdir(relloc):
                    os.makedirs(relloc)
                abs_gcno = os.path.join(self.repository_rootdir, gcno_file)
                ERROR_HANDLER.assert_true(os.path.isfile(abs_gcno), \
                                    "gcno file missing after build", __file__)
                shutil.copy2(abs_gcno, os.path.join(gc_files_dir, gcno_file))
                # FIXME: For now, we store all gcda in th same directory. 
                # FIXME: To ensure that coverage is generated with gcov when 
                # FIXME: The source is not in rootdir (example of coreutils...)
                gcno_dir, gcno_base = os.path.split(\
                                                os.path.normpath(gcno_file))
                gcno_in_top = os.path.join(gc_files_dir, gcno_base)
                #logging.debug("gcno_instr: '{}' '{}' '{}'".format(gcno_file, \
                                                    #gcno_dir, gcno_in_top))
                if gcno_dir != '':
                    # The gcno file must not be in the top gcno_gcda folder
                    ERROR_HANDLER.assert_true(not os.path.isfile(gcno_in_top),\
                            "Overriding of gcno ({}). Source name clash?"\
                                                .format(gcno_file), __file__)
                    shutil.copy2(abs_gcno, gcno_in_top)
                #~ FIXME end

            self._copy_from_repo(rel_path_map)
            return DefaultCallbackObject.after_command(self)
        #~ def after_command()
    #~ class InstrumentCallbackObject

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in ('gcc', 'gcov'):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[0]):
                return False
        return True
    #~ def installed()

    @classmethod
    def _get_meta_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in same file
            :return: list of citeria
        """
        return [
                TestCriteria.STATEMENT_COVERAGE,
                TestCriteria.BRANCH_COVERAGE,
                TestCriteria.FUNCTION_COVERAGE,
               ]
    #~ def _get_meta_instrumentation_criteria()

    @classmethod
    def _get_separated_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in different files
            :return: list of citeria
        """
        return []
    #~ def _get_separated_instrumentation_criteria()

    def get_instrumented_executable_paths_map(self, enabled_criteria):
        using_gdb_wrapper = self.driver_config.get_use_gdb_wrapper()
        if using_gdb_wrapper:
            # use wrapper if gdb is installed
            using_gdb_wrapper = DriversUtils.check_tool(prog='gdb', \
                                                    args_list=['--version'], \
                                                    expected_exit_codes=[0])
            
            if using_gdb_wrapper:
                # XXX: Docker has issues running programs in Docker, 
                # unless the container is ran with the following arguments:
                #
                # docker run --cap-add=SYS_PTRACE \
                #               --security-opt seccomp=unconfined ...
                #
                # We check that it is fine by testing on echo
                ret, o_e, _ = DriversUtils.execute_and_get_retcode_out_err(
                                        prog='gdb', \
                                        args_list=['--batch-silent', \
                                                    '--quiet', 
                                                    '--return-child-result', 
                                                    '-ex', 'run', 
                                                    '--args', 'echo'], \
                                        )#out_on=False, err_on=False)
                using_gdb_wrapper = (ret == 0)
                if not using_gdb_wrapper:
                    logging.warning("use gdb is enabled but call to gdb fails"
                                " (retcode {}) with msg: {}".format(ret, o_e))
            else:
                logging.warning("use gdb is enabled but gdb is not installed")

        crit_to_exes_map = {}
        obj = common_fs.loadJSON(self.instrumentation_details)
        #exes = [p for _, p in list(obj.items())]
        for name in obj:
            obj[name] = os.path.join(self.instrumented_code_storage_dir, \
                                                                    obj[name])
            
        if using_gdb_wrapper:
            # Using GDB WRAPPER
            with open(self.gcov_gdb_wrapper_template) as f:
                template_str = f.read()
            single_exe = obj[list(obj)[0]]
            template_str = template_str.replace(\
                                'MUTERIA_GCOV_PROGRAMEXE_PATHNAME', single_exe)
            with open(self.gcov_gdb_wrapper_sh, 'w') as f:
                f.write(template_str)
            shutil.copymode(single_exe, self.gcov_gdb_wrapper_sh)
            obj[list(obj)[0]] = self.gcov_gdb_wrapper_sh

        exes_map = obj
        for criterion in enabled_criteria:
            crit_to_exes_map[criterion] = exes_map
            
        #logging.debug("DBG: {} {} {}".format(\
        #                  "Using gdb wrapper is {}.".format(using_gdb_wrapper), \
        #                  "crit_to_exe_map is {}.".format(crit_to_exes_map), \
        #                  "wraper path is {}!".format(self.gcov_gdb_wrapper_sh)))
            
        return crit_to_exes_map
    #~ def get_instrumented_executable_paths_map()

    def get_criterion_info_object(self, criterion):
        return None
    #~ def get_criterion_info_object(self, criterion)

    def _get_criterion_element_executable_path(self, criterion, element_id):
        ERROR_HANDLER.error_exit("not applicable for gcov", __file__)
    #~ def _get_criterion_element_executable_path

    def _get_criterion_element_environment_vars(self, criterion, element_id):
        '''
            return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        ERROR_HANDLER.error_exit("not applicable for gcov", __file__)
    #~ def _get_criterion_element_environment_vars()

    def _get_criteria_environment_vars(self, result_dir_tmp, enabled_criteria):
        '''
        return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        return {e:None for e in enabled_criteria}
    #~ def _get_criteria_environment_vars()

    def _collect_temporary_coverage_data(self, criteria_name_list, \
                                            test_execution_verdict, \
                                            used_environment_vars, \
                                            result_dir_tmp, \
                                            testcase):
        ''' get gcov files from gcda files into result_dir_tmp
        '''
        prog = 'gcov'
        if self.custom_binary_dir is not None:
            prog = os.path.join(self.custom_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                                        .format(os.path.basename(prog), \
                                            self.custom_binary_dir), __file__)

        cov2flags = {
                    TestCriteria.STATEMENT_COVERAGE: [],
                    TestCriteria.BRANCH_COVERAGE: ['-b', '-c'],
                    TestCriteria.FUNCTION_COVERAGE: ['-f'],
                }

        args_list = []
        for criterion in criteria_name_list:
            args_list += cov2flags[criterion]

        gcda_files = self._get_gcda_list()

        raw_filename_list = [os.path.splitext(f)[0] for f in gcda_files]
        args_list += raw_filename_list
        
        if len(gcda_files) > 0:
            # TODO: When gcov generate coverage for different files with
            # same name filename but located at diferent dir. Avoid override.
            # Go where the gcov will be looked for
            cwd = os.getcwd()
            os.chdir(self.gc_files_dir)

            # collect gcda (gcno)
            r, _, err_str = DriversUtils.execute_and_get_retcode_out_err(\
                                        prog=prog, \
                                        args_list=args_list, out_on=False, \
                                        err_on=True, merge_err_to_out=False)

            os.chdir(cwd)
            
            if r != 0: # or err_str:
                ERROR_HANDLER.error_exit("Program {} {}.".format(prog,\
                        'collecting coverage is problematic. ')+
                        "The error msg is {}. \nThe command:\n{}".format(\
                                err_str, " ".join([prog]+args_list)), __file__)
            
            dot_gcov_file_list = self._get_gcov_list()
            # Sources of interest
            _, src_map = self.code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()
            base_dot_gcov = [os.path.basename(f) for f in dot_gcov_file_list]
            base_raw = [os.path.basename(f) for f in raw_filename_list]
            interest = [os.path.basename(s) for s,o in src_map.items()]
            interest = [s+'.gcov' for s in interest if os.path.splitext(s)[0] \
                                                                in base_raw]
            missed = set(interest) - set(base_dot_gcov)

            # FIXME: Check for partial failure (compare number of gcno and gcov)
            ERROR_HANDLER.assert_true(len(missed) == 0,
                    "{} did not generate the '.gcov' files {}.".format(\
                        prog, missed)+ "\nIts stderr is {}.\n CMD: {}".format(\
                            err_str, " ".join([prog]+args_list)), __file__)

            # delete gcda
            for gcda_f in gcda_files:
                os.remove(gcda_f)
        else:
            if not self.driver_config.get_allow_missing_coverage():
                ERROR_HANDLER.error_exit(\
                    "Testcase '{}' did not generate gcda, {}".format(\
                        testcase, "when allow missing coverage is disabled"))
            dot_gcov_file_list = []

        common_fs.dumpJSON(dot_gcov_file_list, \
                                os.path.join(result_dir_tmp,\
                                                self.gcov_files_list_filename))
    #~ def _collect_temporary_coverage_data()

    def _extract_coverage_data_of_a_test(self, enabled_criteria, \
                                    test_execution_verdict, result_dir_tmp):
        ''' read json files and extract data
            return: the dict of criteria with covering count
            # TODO: Restrict to returning coverage of specified headers files
        '''
        gcov_list = common_fs.loadJSON(os.path.join(result_dir_tmp,\
                                                self.gcov_files_list_filename))
        
        res = {c: {} for c in enabled_criteria}

        func_cov = None
        branch_cov = None
        statement_cov = {}
        if TestCriteria.FUNCTION_COVERAGE in enabled_criteria:
            func_cov = res[TestCriteria.FUNCTION_COVERAGE]
        if TestCriteria.BRANCH_COVERAGE in enabled_criteria:
            branch_cov = res[TestCriteria.BRANCH_COVERAGE]
        if TestCriteria.STATEMENT_COVERAGE in enabled_criteria:
            statement_cov = res[TestCriteria.STATEMENT_COVERAGE]

        # Sources of interest
        _, src_map = self.code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()
        
        #logging.debug("gcov_list: {}".format(gcov_list))
        
        for gcov_file in gcov_list:
            with open(gcov_file) as fp:
                last_line = None
                src_file = None
                for raw_line in fp:
                    line = raw_line.strip()
                    col_split = [v.strip() for v in line.split(':')]

                    if len(col_split) > 2 and col_split[1] == '0':
                        # preamble
                        if col_split[2] == "Source":
                            src_file = os.path.normpath(col_split[3])
                            if not src_file in src_map:
                                s_cand = None
                                for s in src_map.keys():
                                    if s.endswith(os.sep+src_file):
                                        ERROR_HANDLER.assert_true(s_cand is None,\
                                              "multiple candidate, maybe same "+\
                                              "source name in different dirs "+\
                                              "for source {}".format(src_file), \
                                                                        __file__)
                                        s_cand = s
                                if s_cand is None:
                                    # If src file does not represent (not equal
                                    # and not relatively equal, for case where
                                    # build happens not in rootdir),
                                    # src not in considered
                                    break
                                else:
                                    # ensure compatibility in matrices
                                    src_file = s_cand
                    elif line.startswith("function "):
                        # match function
                        parts = line.split()
                        ident = DriversUtils.make_meta_element(parts[1], \
                                                                    src_file)
                        func_cov[ident] = int(parts[3])
                    elif line.startswith("branch "):
                        # match branch
                        parts = line.split()
                        ident = DriversUtils.make_meta_element(parts[1], \
                                                                    last_line)
                        if parts[2:4] == ['never', 'executed']:
                            branch_cov[ident] = 0
                        else:
                            branch_cov[ident] = int(parts[3])

                    elif len(col_split) > 2 and \
                                            re.match(r"^\d+$", col_split[1]):
                        # match line
                        if col_split[0] == '-':
                            continue
                        last_line = DriversUtils.make_meta_element(\
                                                        col_split[1], src_file)
                        if col_split[0] in ('#####', '====='):
                            exec_count = 0
                        else:
                            exec_count = \
                                    int(re.findall(r'^\d+', col_split[0])[0])
                        statement_cov[last_line] = exec_count

        # delete gcov files
        for gcov_f in self._get_gcov_list():
            os.remove(gcov_f)

        #logging.debug("gcov res is: {}".format(res))
        
        return res
    #~ def _extract_coverage_data_of_a_test()

    def _do_instrument_code (self, exe_path_map, \
                                        code_builds_factory, \
                                        enabled_criteria, parallel_count=1):
        # Setup
        if os.path.isdir(self.instrumented_code_storage_dir):
            shutil.rmtree(self.instrumented_code_storage_dir)
        os.mkdir(self.instrumented_code_storage_dir)
        if os.path.isdir(self.gc_files_dir):
            try:
                shutil.rmtree(self.gc_files_dir)
            except PermissionError:
                self._dir_chmod777(self.gc_files_dir)
                shutil.rmtree(self.gc_files_dir)
        os.mkdir(self.gc_files_dir, mode=0o777)

        prog = 'gcc'
        if self.custom_binary_dir is not None:
            prog = os.path.join(self.custom_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                                        .format(os.path.basename(prog), \
                                            self.custom_binary_dir), __file__)

        flags = ['-g', '--coverage', '-fprofile-dir='+self.gc_files_dir, '-O0']
        additionals = ["-fkeep-inline-functions"]
        #additionals.append('-fprofile-abs-path')
        #additionals.append('-fprofile-prefix-path='+code_builds_factory\
        #                        .repository_manager.get_repository_dir_path())
        
        # get gcc version
        ret, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                            prog, args_list=['-dumpversion'])
        ERROR_HANDLER.assert_true(ret == 0, "'gcc -dumpversion' failed'")
        
        # if version > 6.5
        if int(out.split('.')[0]) >= 6:
            if int(out.split('.')[0]) > 6 or int(out.split('.')[1]) > 5:
                additionals += ["-fkeep-static-functions"]
        
        flags += additionals
        
        rel_rel_path_map = {}
        rel_abs_path_map = {}
        exes, _ = code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()
        ERROR_HANDLER.assert_true(len(exes) == 1, \
                                "Only single exe supported in GCOV", __file__)
        for exe in exes:
            filename = os.path.basename(exe)
            rel_rel_path_map[exe] = filename
            rel_abs_path_map[exe] = os.path.join(\
                                self.instrumented_code_storage_dir, filename)

        self.instrument_callback_obj = self.InstrumentCallbackObject()
        self.instrument_callback_obj.set_post_callback_args(\
                                        (self.gc_files_dir, rel_abs_path_map))
        pre_ret, ret, post_ret = code_builds_factory.transform_src_into_dest(\
                        src_fmt=CodeFormats.C_SOURCE,\
                        dest_fmt=CodeFormats.NATIVE_CODE,\
                        src_dest_files_paths_map=None,\
                        compiler=prog, flags_list=flags, clean_tmp=True, \
                        reconfigure=True, \
                        callback_object=self.instrument_callback_obj)
        
        # Check
        if ret == common_mix.GlobalConstants.COMMAND_FAILURE:
            ERROR_HANDLER.error_exit("Program {} {}.".format(prog,\
                                        'built problematic'), __file__)

        # write down the rel_path_map
        ERROR_HANDLER.assert_true(not os.path.isfile(\
                self.instrumentation_details), "must not exist here", __file__)
        common_fs.dumpJSON(rel_rel_path_map, self.instrumentation_details)

    #~ def _do_instrument_code()
#~ class CriteriaToolGCov




from muteria.drivers.criteria.tools_by_languages.c.gpmutation.gpmutation \
                                                import CriteriaToolGPMutation 


StaticCriteriaTool = CriteriaToolGPMutation




from __future__ import print_function
import os
import sys
import glob
import shutil
import logging
import resource
import random

import numpy as np

import muteria.common.fs as common_fs
import muteria.common.mix as common_mix
import muteria.common.matrices as common_matrices

import muteria.drivers.criteria as criteria
import muteria.controller.explorer as fd_structure

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.drivers.testgeneration.base_testcasetool import BaseTestcaseTool
from muteria.drivers.testgeneration.testcases_info import TestcasesInfoObject
from muteria.drivers import DriversUtils
from muteria.drivers.testgeneration.testcase_formats.ktest.ktest \
                                                        import KTestTestFormat
from muteria.drivers.testgeneration.tools_by_languages.c.klee.klee \
                                                    import TestcasesToolKlee

from muteria.drivers.testgeneration.tools_by_languages.c.semu.driver_config \
                                        import DriverConfigSemu, MetaMuSource

ERROR_HANDLER = common_mix.ErrorHandler

class TestcasesToolSemu(TestcasesToolKlee):

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in ('klee-semu',):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[0,1]):
                return False

        return KTestTestFormat.installed(custom_binary_dir=custom_binary_dir)
    #~ def installed()

    def __init__(self, *args, **kwargs):
        TestcasesToolKlee.__init__(self, *args, **kwargs)
        
        self.driver_config = None
        if self.config.get_tool_user_custom() is not None:
            self.driver_config = \
                            self.config.get_tool_user_custom().DRIVER_CONFIG
        if self.driver_config is None:
            self.driver_config = DriverConfigSemu()
        else:
            ERROR_HANDLER.assert_true(isinstance(self.driver_config, \
                                            DriverConfigSemu),\
                                            "invalid driver config", __file__)

        self.cand_muts_file = os.path.join(self.tests_working_dir, \
                                                        "cand_muts_file.txt")
        self.sm_mat_file = self.head_explorer.get_file_pathname(\
                            fd_structure.CRITERIA_MATRIX[criteria.TestCriteria\
                                                            .STRONG_MUTATION])
        self.mutants_by_funcs = None
    #~ def __init__()

    # SHADOW override
    def _get_default_params(self):
        bool_params = {
            '-ignore-solver-failures': None,
            '-allow-external-sym-calls': True, #None,
            '-posix-runtime': True, #None,
            '-dump-states-on-halt': True, #None,
            '-only-output-states-covering-new': None,
            '-use-cex-cache': True,
          
            '-only-replay-seeds': None, # Useful to set this is using seeds
            # SEMu 
            '-semu-no-state-difference': None,
            '-semu-MDO-propagation-selection-strategy': None,
            '-semu-use-basicblock-for-distance': None,
            '-semu-forkprocessfor-segv-externalcalls': True,
            '-semu-testsgen-only-for-critical-diffs': None,
            '-semu-no-environment-output-diff': None,
            '-semu-use-only-multi-branching-for-depth': None,
            '-semu-disable-post-mutation-check': None,
            '-semu-no-error-on-memory-limit': None,
        }
        key_val_params = {
            '-output-dir': self.tests_storage_dir,
            '-solver-backend': 'z3',
            '-max-solver-time': '300',
            '-search': 'bfs',
            '-max-memory': None,
            '-max-time': self.config.TEST_GENERATION_MAXTIME,
            '-libc': 'uclibc',
            '-max-sym-array-size': '4096',
            '-max-instruction-time': '10.',
          
            '-'+self.SEED_DIR_ARG_NAME: None, 
            # SEMu 
            '-semu-checkpoint-window': '0', #None,
            '-semu-minimum-propagation-depth': '2', # None,
            '-semu-propagation-proportion': '0.25', #None,
            '-semu-precondition-length': '0', #None,
            '-semu-max-total-tests-gen': None,
            '-semu-number-of-tests-per-mutant': '5', # None,
            '-semu-loop-break-delay': '120.0',
        }
        key_val_params.update({
                                })
        # XXX: set muts cand list
        if os.path.isfile(self.sm_mat_file):
            key_val_params['-semu-candidate-mutants-list-file'] = \
                                                        self.cand_muts_file
        return bool_params, key_val_params
    #~ def _get_default_params()
    
    def _handle_backward_compatibility_with_semu (self, args):
        to_replace_map = {
            '-semu-mutant-max-fork': '-semu-checkpoint-window',
            '--semu-mutant-max-fork': '--semu-checkpoint-window',
            '-semu-mutant-state-continue-proba': \
                                            '-semu-propagation-proportion',
            '--semu-mutant-state-continue-proba': \
                                           '--semu-propagation-proportion',
            '-semu-continue-mindist-out-heuristic': \
                                '-semu-MDO-propagation-selection-strategy',
            '--semu-continue-mindist-out-heuristic': \
                                '--semu-MDO-propagation-selection-strategy',
            '-semu-checknum-before-testgen-for-discarded': \
                                        '-semu-minimum-propagation-depth',
            '--semu-checknum-before-testgen-for-discarded': \
                                        '--semu-minimum-propagation-depth',
            '-semu-disable-statediff-in-testgen': \
                                               '-semu-no-state-difference',
            '--semu-disable-statediff-in-testgen': \
                                            '--semu-no-state-difference',
            '-semu-max-tests-gen-per-mutant': \
                                        '-semu-number-of-tests-per-mutant',
            '--semu-max-tests-gen-per-mutant': \
                                        '--semu-number-of-tests-per-mutant'
        }
        
        # Direct replace
        old_version = False
        for pos, val in enumerate(args):
            for match, replace in to_replace_map.items():
                if val == match:
                    args[pos] = replace
                    old_version = True
                    break
                    
        # negation (Now out env is set by default)
        found_pos = None
        for match in ['-semu-consider-outenv-for-diffs', \
                                    '--semu-consider-outenv-for-diffs']:
            try:
                pos = args.index(match)
                # present
                ERROR_HANDLER.assert_true (found_pos is None, \
                                 "Multiple occurence of "
                           "'-semu-consider-outenv-for-diffs'", __file__)
                found_pos = pos
            except ValueError:
                # abscent
                pass
        if old_version:
            if found_pos is None:
                # insert no outenv
                args.insert(0, '--semu-no-environment-output-diff')
            else:
                del args[found_pos]
    #~ def _handle_backward_compatibility_with_semu()
    
    def _call_generation_run(self, runtool, args):
        # Handle support for alod version of SEMu
        self._handle_backward_compatibility_with_semu (args)
        
        # If seed-dir is set, ensure that only-replay-seeds is set 
        # (semu requires it for now)
        seed_dir_key = self.SEED_DIR_ARG_NAME
        only_replay_seeds_flag = '-only-replay-seeds'
        if self.get_value_in_arglist(args, seed_dir_key) is not None:
            if only_replay_seeds_flag not in args and \
                                  '-'+only_replay_seeds_flag not in args: 
                logging.warning("SEMu requires '-only-replay-seeds' flag to"
                                " be set when using seeds. Muteria set it"
                                " autmatically")
                args.insert(0, only_replay_seeds_flag)
                
        # use mutants_by_funcs to reorganize target mutants for scalability

        if os.path.isfile(self.cand_muts_file):
            max_mutant_count_per_cluster = \
                            self.driver_config.get_max_mutant_count_per_cluster()

            cand_mut_file_bak = self.cand_muts_file + '.bak'
            mut_list = []
            with open(self.cand_muts_file) as f:
                for m in f:
                    m = m.strip()
                    ERROR_HANDLER.assert_true(m.isdigit(), "Invalid mutant ID", \
                                                                        __file__)
                    mut_list.append(m)
            if self.mutants_by_funcs is None:
                random.shuffle(mut_list)
            else:
                # make the mutants to be localized in same function as much as
                # possible
                # XXX Here the mutant ID are NOT meta but simple IDs
                this_mutants_by_funcs = {}
                mut_to_func = {}
                for f, m_set in self.mutants_by_funcs.items():
                    this_mutants_by_funcs[f] = m_set & set(mut_list)
                    for m in this_mutants_by_funcs[f]:
                        mut_to_func[m] = f
                func_to_rank_dec = list(this_mutants_by_funcs.keys())
                func_to_rank_dec.sort(key=lambda x: len(this_mutants_by_funcs[x]),\
                                                                    reverse=True)
                func_to_rank_dec = {f: r for r, f in enumerate(func_to_rank_dec)}
                mut_list.sort(key=lambda x: func_to_rank_dec[mut_to_func[x]])
            nclust = int(len(mut_list) / max_mutant_count_per_cluster)
            if len(mut_list) != max_mutant_count_per_cluster * nclust:
                nclust += 1
            clusters = np.array_split(mut_list, nclust)

            # update max-time
            if len(clusters) > 1:
                cur_max_time = float(self.get_value_in_arglist(args, 'max-time'))
                self.set_value_in_arglist(args, 'max-time', \
                                        str(max(1, cur_max_time / len(clusters))))
            
            shutil.move(self.cand_muts_file, cand_mut_file_bak)

            c_dirs = []
            for c_id, clust in enumerate(clusters):
                logging.debug("SEMU: targeting mutant cluster {}/{} ...".format(\
                                                            c_id+1, len(clusters)))
                with open(self.cand_muts_file, 'w') as f:
                    for m in clust:
                        f.write(m+'\n')

                super(TestcasesToolSemu, self)._call_generation_run(runtool, args)
                
                c_dir = os.path.join(os.path.dirname(self.tests_storage_dir), \
                                                                        str(c_id))
                shutil.move(self.tests_storage_dir, c_dir)
                c_dirs.append(c_dir)

            os.mkdir(self.tests_storage_dir)
            for c_dir in c_dirs:
                shutil.move(c_dir, self.tests_storage_dir)

            shutil.move(cand_mut_file_bak, self.cand_muts_file)
        else:
            super(TestcasesToolSemu, self)._call_generation_run(runtool, args)
    #~ def _call_generation_run()

    def _get_tool_name(self):
        return 'klee-semu'
    #~ def _get_tool_name()
    
    def _get_compile_flags_list(self):
        return ['-DMUTERIA_FOR_SEMU_TEST_GENERATION']
    #~ def _get_compile_flags_list()

    def _get_input_bitcode_file(self, code_builds_factory, rel_path_map, \
                                                meta_criteria_tool_obj=None):
        meta_mu_src = self.driver_config.get_meta_mutant_source()

        # XXX Case of manual annotation
        if meta_mu_src == MetaMuSource.ANNOTATION:
            with open(self.cand_muts_file, 'w') as f:
                # Single mutant (id 1, corresponding to old version)
                f.write(str(1)+'\n')
            return super(TestcasesToolSemu, self)._get_input_bitcode_file(\
                                        code_builds_factory, rel_path_map, \
                                meta_criteria_tool_obj=meta_criteria_tool_obj)
        if type(meta_mu_src) == str:
            # XXX: The actual path to the meta is specified
            return meta_mu_src
        
        # XXX: Case of other mutation tools like Mart
        # get the meta criterion file from MART or any compatible tool.
        mutant_gen_tool_name = meta_mu_src.get_field_value()
        mut_tool_alias_to_obj = \
                            meta_criteria_tool_obj.get_criteria_tools_by_name(\
                                                        mutant_gen_tool_name)

        if len(mut_tool_alias_to_obj) == 0:
            logging.warning(\
                'SEMu requires {} to generate mutants but none used'.format(\
                                                        mutant_gen_tool_name))

        ERROR_HANDLER.assert_true(len(mut_tool_alias_to_obj) == 1, \
                                "SEMu supports tests generation from"
                                "a single .bc file for now (todo).", __file__)

        t_alias2metamu_bc = {}
        t_alias2mutantInfos = {}
        for alias, obj in mut_tool_alias_to_obj.items():
            dest_bc = rel_path_map[list(rel_path_map)[0]]+'.bc'
            shutil.copy2(obj.get_test_gen_metamutant_bc(), dest_bc)
            t_alias2metamu_bc[alias] = dest_bc
            t_alias2mutantInfos[alias] = obj.get_criterion_info_object(None)

        # XXX: get mutants ids by functions
        self.mutants_by_funcs = {}
        single_alias = list(t_alias2mutantInfos)[0]
        single_tool_obj = t_alias2mutantInfos[single_alias]

        cand_muts = list(single_tool_obj.get_elements_list())
        
        for mut in single_tool_obj.get_elements_list():
            func = single_tool_obj.get_element_data(mut)[\
                                                        'mutant_function_name']
            #meta_mut = DriversUtils.make_meta_element(mut, single_alias)
            if func not in self.mutants_by_funcs:
                self.mutants_by_funcs[func] = set()
            self.mutants_by_funcs[func].add(mut) #meta_mut)

        # XXX: get candidate mutants list
        if self.driver_config.get_target_only_live_mutants() \
                                        and os.path.isfile(self.sm_mat_file):
            sm_mat = common_matrices.ExecutionMatrix(\
                                                filename=self.sm_mat_file)
            mut2killing_tests = sm_mat.query_active_columns_of_rows()
            alive_muts = [m for m, k_t in mut2killing_tests.items() \
                                                        if len(k_t) == 0]
            cand_muts = []
            for meta_m in alive_muts:
                t_alias, m = DriversUtils.reverse_meta_element(meta_m)
                if t_alias in t_alias2metamu_bc: # There is a single one
                    cand_muts.append(m)

        with open(self.cand_muts_file, 'w') as f:
            for m in cand_muts:
                f.write(str(m)+'\n')

        return t_alias2metamu_bc[list(t_alias2metamu_bc)[0]]
    #~ def _get_input_bitcode_file()

    @staticmethod
    def _get_generation_time_of_test(test, test_top_dir):
        """ extract the generation timestamp of a test. the test and 
            its top location are specified
        """
        test_path = os.path.join(test_top_dir, test)
        # Seach the test and get the time. 
        # If dup, get the time of the duplicate test
        # 1. If test_path exist, is is not dup
        to_search = None
        if os.path.isfile(test_path):
            to_search = test_path
        else:
            # seach the test it is dup of
            kepttest2duptest_map = common_fs.loadJSON(\
                                              self.keptktest2dupktests)
            ERROR_HANDLER.asser_true(test not in kepttest2duptest_map, \
                            "Test file not existing but in kept (BUG)", \
                                                                __file__)
            for kept, dup_list in kepttest2duptest_map.items():
                if test in dup_list:
                    to_search = os.path.join(test_top_dir, kept)
        
        ERROR_HANDLER.assert_true(to_search is not None, \
                        "test not found in dup and inexistant", __file__)
        # 2. search for the test and its time
        ktestlists = glob.glob(os.path.join(os.path.dirname(to_search), \
                                              "mutant-[0-9]*.ktestlist"))
        gen_time = None
        for klist in ktestlists:
            df = common_fs.loadCSV(klist,  separator=",")
            val = df.loc[df.ktest==os.path.basename(to_search), \
                                                'ellapsedTime(s)'].values
            if len(val) == 1:
                gen_time = val[0]
                break
            ERROR_HANDLER.assert_true(len(val) == 0, "PB in ktestlist " + \
                  str(klist) + ". ktest appearing multiple times", __file__)
        ERROR_HANDLER.assert_true(gen_time is not None, \
                                   "test not found in ktestlists", __file__)
        return gen_time
    #~ def _get_generation_time_of_test()
    
    def fdupeGeneratedTest (self, mfi_ktests_dir_top, mfi_ktests_dir, \
                                                semuoutputs, seeds_dir=None):
        # Implement and use (see run.py in Semu Analysis)
        # TODO, implement this
        ERROR_HANDLER.error_exit("TODO: implement fdupesGeneratedTest")
    #~ def fdupeGeneratedTest ()

    def requires_criteria_instrumented(self):
        return self.driver_config.requires_criteria_instrumented()
    #~ def requires_criteria_instrumented()
#~ class TestcasesToolSemu



from __future__ import print_function

import os
import sys
import re
import shutil
import imp
import logging
import filecmp
from distutils.spawn import find_executable

import muteria.common.mix as common_mix

from muteria.drivers import DriversUtils

ERROR_HANDLER = common_mix.ErrorHandler

class KTestTestFormat(object):
    
    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in (cls.tool,):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[1]):
                return False
        if not DriversUtils.check_tool(prog=cls.stdbuf, args_list=['--version'],\
                                                    expected_exit_codes=[0]):
                return False
        return True
    #~ def installed()

    @classmethod
    def get_test_replay_tool(cls, custom_replay_tool_binary_dir=None):
        if custom_replay_tool_binary_dir is None:
            kr_file = find_executable(cls.tool)
            ERROR_HANDLER.assert_true(kr_file is not None, \
                        "Could not fine test replay tool on path", __file__)
        else:
            kr_file = os.path.join(custom_replay_tool_binary_dir, cls.tool)
            ERROR_HANDLER.assert_true(os.path.isfile(kr_file), \
                                "replay tool not found in custom_binary_dir", \
                                                                    __file__)
        return kr_file
    #~ def get_test_replay_tool()

    @classmethod
    def _get_replay_prog_args(cls, executable_file, test_file, \
                                        custom_replay_tool_binary_dir=None):
        prog = cls.tool
        if custom_replay_tool_binary_dir is not None:
            prog = os.path.join(custom_replay_tool_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                            .format(cls.tool, custom_replay_tool_binary_dir), \
                                                                    __file__)

        args = [executable_file, test_file]
        return prog, args
    #~ def _get_replay_prog_args()

    @classmethod
    def execute_test(cls, executable_file, test_file, env_vars, stdin=None, \
                                        must_exist_dir_list=None, \
                                        timeout=None, collected_output=None, \
                                        custom_replay_tool_binary_dir=None):

        prog, args = cls._get_replay_prog_args(executable_file, test_file, \
                                                custom_replay_tool_binary_dir)

        # klee-replay may create files or dir. in KLEE version with LLVM-3.4,
        # those are created in a temporary dir set as <cwd>.temps
        # XXX XXX. make sure each test has its own
        test_work_dir = test_file+".execdir"
        klee_replay_temps = test_work_dir + '.temps'
        for d in (test_work_dir, klee_replay_temps):
            if os.path.isdir(d):
                try:
                    shutil.rmtree(d)
                except PermissionError:
                    cls._dir_chmod777(d)
                    shutil.rmtree(d)
                
        if not os.path.isdir(test_work_dir):
            os.mkdir(test_work_dir)
            
        if must_exist_dir_list is not None:
            for d in must_exist_dir_list:
                td = os.path.join(test_work_dir, d)
                if not os.path.isdir(td):
                    os.makedirs(td)

        # XXX Execution setup
        tmp_env = os.environ.copy()
        if env_vars is not None:
            #for e, v in env_vars.items():
            #    tmp_env[e] = v
            tmp_env.update(env_vars)

        timeout_return_codes = cls.timedout_retcodes + \
                                        DriversUtils.EXEC_TIMED_OUT_RET_CODE

        if timeout is not None:
            tmp_env['KLEE_REPLAY_TIMEOUT'] = str(timeout)
            kt_over = 10 # 1second
            timeout += kt_over
        else:
            # DBG
            logging.warning("@KTEST: calling ktest execution without timeout.")

        # XXX Get the parsing regexes to use
        retcode, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                prog=prog, args_list=['--help'], \
                                merge_err_to_out=True)
        clean_regex, status_regex = cls._get_regexes(out, \
                                                        clean_everything=True)
        
        # XXX Execute the ktest
        #logging.debug("DBG: test_work_dir is {}. its content is {}".format(
        #                    test_work_dir, list(os.listdir(test_work_dir))))
        #if collected_output is not None:
        
        # XXX: Use stdbuf to line buffer the stderr to avoid mixing or 
        # err between klee-replay and the executd prog
        use_stdbuf = True
        if use_stdbuf:
            args = ["--output=L", "--error=L", prog] + args
            prog = "stdbuf"
            # TODO: check that stdbuf is installed
            
        retcode, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                prog=prog, args_list=args, env=tmp_env, \
                                stdin=stdin, \
                                timeout=timeout, timeout_grace_period=5, \
                                merge_err_to_out=True, cwd=test_work_dir)
        retcode, out, exit_status = cls._remove_output_noise(retcode, out, \
                                                  clean_regex, status_regex)
        # In klee-replay, when exit_status here is not None, retcode is 0
        # When there is an issue, like timeout, exit_status is None and
        # retcode has the ode of the issue 
        if exit_status is None:
            exit_status = retcode
        if collected_output is not None:
            collected_output.extend((exit_status, out, \
                                     (retcode in timeout_return_codes or \
                            retcode in DriversUtils.EXEC_TIMED_OUT_RET_CODE)))
        #else:
        #    retcode, out, err = DriversUtils.execute_and_get_retcode_out_err(\
        #                        prog=prog, args_list=args, env=tmp_env, \
        #                        timeout=timeout, timeout_grace_period=5, \
        #                                        out_on=False, err_on=False, \
        #                                        cwd=test_work_dir)

        # XXX: Go back to previous CWD
        for d in (test_work_dir, klee_replay_temps):
            if os.path.isdir(d):
                try:
                    shutil.rmtree(d)
                except PermissionError:
                    cls._dir_chmod777(d)
                    shutil.rmtree(d)
                
        #if must_exist_dir_list is not None:
        #    try:
        #        shutil.rmtree(test_work_dir)
        #    except PermissionError:
        #        cls._dir_chmod777(test_work_dir)
        #        shutil.rmtree(test_work_dir)

        if retcode in timeout_return_codes + \
                                    DriversUtils.EXEC_SEGFAULT_OUT_RET_CODE:
            verdict = common_mix.GlobalConstants.FAIL_TEST_VERDICT
        else:
            verdict = common_mix.GlobalConstants.PASS_TEST_VERDICT

        return verdict
    #~ def execute_test()

    @staticmethod
    def _dir_chmod777(dirpath):
        try:
            for root_, dirs_, files_ in os.walk(dirpath):
                for sub_d in dirs_:
                    if os.path.isdir(os.path.join(root_, sub_d)):
                        os.chmod(os.path.join(root_, sub_d), 0o777)
                for f_ in files_:
                    if os.path.isfile(os.path.join(root_, f_)):
                        os.chmod(os.path.join(root_, f_), 0o777)
        except PermissionError:
            ret,out,_ = DriversUtils.execute_and_get_retcode_out_err('sudo', \
                                    args_list=['chmod', '777', '-R', dirpath])
            ERROR_HANDLER.assert_true(ret == 0, \
                        "'sudo chmod 777 -R "+dirpath+"' failed (returned "+\
                                        str(ret)+"), error: "+out, __file__)
    #~ def _dir_chmod777()
    
    @classmethod
    def get_replay_test_wrapper_str(cls, exe_env_var, ktest_env_var, \
                                                            timeout_env_var, \
                                        custom_replay_tool_binary_dir=None):
        # XXX: This is used by shadow to replay generated tests through
        # executing base dev test used to generate them.
        # TODO: Let shadow store the mapping between dev test and gen test
        # TODO: Use the mapping to run tests using the returned wrapper by this function
        prog, args = cls._get_replay_prog_args('${}'.format(exe_env_var), \
                                                '${}'.format(ktest_env_var), \
                                               custom_replay_tool_binary_dir)
        
        python_code = ';'.join(['import sys', \
                'from muteria.drivers.testgeneration' \
                    + '.testcase_formats.ktest.ktest import KTestTestFormat', \
            'r, e_s = KTestTestFormat._remove_output_noise(sys.stdin.read())',\
                    'print(r)'])

        bash_timeout_retcode = os.system('timeout 0.1 sleep 1')

        ret_str = "#! /bin/bash\n\n"
        ret_str += "set -u\nset -o pipefail\n\n"
        ret_str += "export KLEE_REPLAY_TIMEOUT={}\n".format(timeout_env_var)
        ret_str += " ".join([prog] + args) + ' 2>&1 | {} -c "{}"\n'.format(\
                                                sys.executable, python_code)
        ret_str += "exit_code=$?\n"
        ret_str += '{} -c "exit(not($exit_code in {}))" && exit_code={}\n'\
                                                .format(sys.executable, \
                                                    cls.timedout_retcodes, \
                                                    bash_timeout_retcode)
        ret_str += "exit $exit_code\n"
        return ret_str
    #~ def get_replay_test_wrapper_str()

    timedout_retcodes = (88,) # taken from klee_replay source code
        
    tool = 'klee-replay'
    stdbuf = 'stdbuf'

    # Newer version (after klee github commit 88bb205)
    # (88bb205e422ee2aaf75594e4e314b21f77f219e3)
    clean_everything_regex_new = re.compile("^KLEE-REPLAY: ")
    clean_part_regex_new = re.compile("(" + "|".join(["^KLEE-REPLAY: NOTE: ",\
                                        "^KLEE-REPLAY: WARNING: ",\
                                        "^KLEE-REPLAY: klee_warning: ",\
                                        "^KLEE-REPLAY: klee_warning_once: ",\
                                        "^KLEE-REPLAY: klee_assume",\
                                        ]) + ")")
    status_regex_new = re.compile(\
                                "^(KLEE-REPLAY: NOTE:\\s+)(EXIT STATUS: .*?)"+\
                                            "(\\s+\\([0-9]+\\s+seconds\\))$")

    # the option "--keep-replay-dir" was added on klee github commit 5b1214a, 
    # right after commit 88bb205
    # So we will use that option to decide whether to use old or new regex
    # Older version (before klee github commit 88bb205)
    clean_everything_regex_old = re.compile("(" + "|".join([\
                        #"^EXIT STATUS: .* \\([0-9]+\\s+seconds\\)$", \
                        #""+tool+": EXIT STATUS: .* \\([0-9]+\\s+seconds\\)$",\
                        ""+tool+": received signal [0-9]+\\s+. "+\
                                        "Killing monitored process\\(es\\)$", \
                        "^note: (pty|pipe) (master|slave): ",\
                        ""+tool+": PTY (MASTER|SLAVE): EXIT STATUS: ", \
                        "^warning: check_file .*: .* "+\
                                        "mismatch: [0-9]+ [vV][sS] [0-9]+$", \
                        "^RUNNING GDB: /usr/bin/gdb --pid [0-9]+ -q --batch", \
                        "^TIMEOUT: ATTEMPTING GDB EXIT$", \
                        #"^ERROR: ", \
                        #"^Error: (executable|chroot:) ", \
                        #"^klee_range(\(|:)", \
                        #"^make_symbolic mismatch, different sizes: ", \
                        #"^WARNING: ", \
                        #"^rootdir: ", \
                        #""+tool+": error: input file ", \
                        ""+tool+": TEST CASE: ", \
                        ""+tool+": ARGS: ", \
                        ]) + ")")
    clean_part_regex_old = re.compile(("(" + "|".join([\
                        #"^EXIT STATUS: .* \\([0-9]+\\s+seconds\\)$", \
                        #""+tool+": EXIT STATUS: .* \\([0-9]+\\s+seconds\\)$", \
                        ""+tool+": received signal [0-9]+\\s+. "+\
                                        "Killing monitored process\\(es\\)$", \
                        "^note: (pty|pipe) (master|slave): ",\
                        ""+tool+": PTY (MASTER|SLAVE): EXIT STATUS: ", \
                        "^warning: check_file .*: .* "+\
                                        "mismatch: [0-9]+ [vV][sS] [0-9]+$", \
                        "^RUNNING GDB: /usr/bin/gdb --pid [0-9]+ -q --batch", \
                        "^TIMEOUT: ATTEMPTING GDB EXIT$", \
                        ]) + ")"))
    status_regex_old = re.compile("^("+tool+\
                    ":\\s+)?(EXIT STATUS: .*?)(\\s+\\([0-9]+\\s+seconds\\))?$")

    @classmethod
    def _get_regexes(cls, out, clean_everything=True):
        if '--keep-replay-dir' in out:
            clean_regex = cls.clean_everything_regex_new if clean_everything \
                                                  else cls.clean_part_regex_new
            status_regex = cls.status_regex_new
        else:
            clean_regex = cls.clean_everything_regex_old if clean_everything \
                                                  else cls.clean_part_regex_old
            status_regex = cls.status_regex_old
        
        return clean_regex, status_regex
    #~ def _get_regexes()
        
    @classmethod
    def _remove_output_noise(cls, retcode, out, clean_regex, status_regex):
        res = []
        
        if len(out) > 0 and out[-1] == '\n':
            out = out[:-1]
            last_char = "\n" 
        else:
            last_char = ""

        # If not None, must be an integer
        exit_status = None

        found_exit_status = False
        for line in out.encode('utf-8', 'backslashreplace').splitlines():
            line = line.decode('utf-8', 'backslashreplace')
            if status_regex.search(line) is not None:
                ERROR_HANDLER.assert_true(not found_exit_status,
                                "Exit status found multiple times in output", \
                                                                      __file__)
                found_exit_status = True
                line = status_regex.sub("\g<2>", line)
                ls = line.split()
                if ls[-2] == 'ABNORMAL':
                    try:
                        exit_status = int(ls[-1])
                    except ValueError:
                        ERROR_HANDLER.error_exit(\
                                    "Invalid exit status {}".format(ls[-1]), \
                                                                 __file__)
                elif ls[-1] == 'OUT' and ls[-2] == 'TIMED':
                    # Case where klee-replay call to gdb fails to attach process
                    if retcode == 0:
                        retcode = cls.timedout_retcodes[0]
                    # klee-replay may pu another exit status
                    found_exit_status = False
                res.append("@MUTERIA.KLEE-REPLAY: "+line)
            elif clean_regex.search(line) is None:
                # None is matched
                res.append(line)

        res = '\n'.join(res) + last_char

        return retcode, res, exit_status
    #~ def _remove_output_noise()

    ktest_extension = '.ktest'
    STDIN_KTEST_DATA_FILE = "muteria-stdin-ktest-data"

    @classmethod
    def ktest_fdupes(cls, *args, custom_replay_tool_binary_dir=None):
        """
        This function computes the fdupes of the klee ktest directories 
        and ktest files given as arguments. 
        It requires that the files and directories passed as arguments exist

        :param *args: each argument is either a file or a directory that exists

        :return: returns two values: 
                - The first is a python list of tuples. 
                    each tuple represents files that are duplicate with each 
                    other and ranked by their age (modified time) the oldest 
                    first (earliest modified to latest modified).
                - The second is the list of files that are not valid
                    ktest files.
        """
        # import ktest
        ktt_dir = os.path.dirname(cls.get_test_replay_tool(
                                        custom_replay_tool_binary_dir=\
                                                custom_replay_tool_binary_dir))
        ktest_tool = imp.load_source("ktest-tool", \
                                        os.path.join(ktt_dir, 'ktest-tool'))

        ret_fdupes = []
        invalid = []
        file_set = set()
        for file_dir in args:
            if os.path.isfile(file_dir):
                file_set.add(file_dir)
            elif os.path.isdir(file_dir):
                # get ktest files recursively
                for root, directories, filenames in os.walk(file_dir):
                    for filename in filenames:
                        if filename.endswith(cls.ktest_extension):
                            file_set.add(os.path.join(root, filename))
            else:
                ERROR_HANDLER.error_exit(\
                        "Invalid file or dir passed (inexistant): "+file_dir, \
                                                                    __file__)

        # apply fdupes: load all ktests and strip the non uniform data 
        # (.bc file used) then compare the remaining data
        kt2used_dat = {}
        kt2stdin = {}
        for kf in file_set:
            try:
                b = ktest_tool.KTest.fromfile(kf)
                if len(b.objects) == 0:
                    invalid.append(kf)
                else:
                    kt2used_dat[kf] = (b.args[1:], b.objects)
            except:
                invalid.append(kf)

            # STDIN
            stdin_file = os.path.join(os.path.dirname(kf), \
                                                    cls.STDIN_KTEST_DATA_FILE)
            if os.path.isfile(stdin_file) \
                                        and os.path.getsize(stdin_file) > 0:
                kt2stdin[kf] = stdin_file
            else:
                kt2stdin[kf] = None

        # do fdupes
        dup_dict = {}
        keys = list(kt2used_dat.keys())
        for ktest_file in keys:
            if ktest_file in kt2used_dat:
                ktest_file_dat = kt2used_dat[ktest_file]
                del kt2used_dat[ktest_file]
                for other_file in kt2used_dat:
                    if kt2used_dat[other_file] == ktest_file_dat:
                        # compare stdin
                        if kt2stdin[ktest_file] == kt2stdin[other_file] or \
                                (kt2stdin[ktest_file] is not None \
                                        and kt2stdin[other_file] is not None \
                                        and filecmp.cmp(kt2stdin[ktest_file], \
                                                        kt2stdin[other_file], 
                                                        shallow=False)):
                            if ktest_file not in dup_dict:
                                dup_dict[ktest_file] = []
                            dup_dict[ktest_file].append(other_file)

                # remove all dupicates from candidates
                if ktest_file in dup_dict:
                    for dup_of_kt_file in dup_dict[ktest_file]:
                        del kt2used_dat[dup_of_kt_file]

        # Finilize
        for ktest_file in dup_dict:
            tmp = [ktest_file] + dup_dict[ktest_file]
            # sort by decreasing modified age
            tmp.sort(key=lambda x: os.path.getmtime(x))
            ret_fdupes.append(tuple(tmp))

        return ret_fdupes, invalid
    #~ def ktest_fdupes()

    @staticmethod
    def get_dir (ktest_fullpath, cand_folders):
        for fold in cand_folders:
            if ktest_fullpath.startswith(fold):
                return fold
        ERROR_HANDLER.error_exit(\
                "Not candidate folder found in {} for ktest {}".format(\
                                cand_folders, ktest_fullpath), __file__)
    #~ def get_dir ()

    @staticmethod
    def cross_tool_fdupes(*testtools):
        clusters = {}
        testdir2ttalias = {}
        for tt in testtools:
            custom_bin = tt.custom_binary_dir
            kt_dir = tt.get_ktests_dir()
            if not os.path.isdir(kt_dir):
                # test generation did not yet take place
                continue
            if custom_bin not in clusters:
                clusters[custom_bin] = []
            clusters[custom_bin].append(kt_dir)
            testdir2ttalias[kt_dir] = tt.config.get_tool_config_alias()

        kepttest2duptest_map = {}
        test2keptdup = {}
        for custom_bin, folders in clusters.items():
            
            tmp_kepttest2duptest_map, tmp_test2keptdup = \
                            KTestTestFormat.cross_folder_fdupes(custom_bin, \
                                                    folders, testdir2ttalias)
            kepttest2duptest_map.update(tmp_kepttest2duptest_map)
            test2keptdup.update(tmp_test2keptdup)
        
        return kepttest2duptest_map, test2keptdup
    #~ def cross_tool_fdupes()

    @staticmethod
    def cross_folder_fdupes(custom_bin, folders, folder2alias):
        dup_list, invalid = KTestTestFormat.ktest_fdupes(*folders, \
                                custom_replay_tool_binary_dir=custom_bin)
        ERROR_HANDLER.assert_true(len(invalid) == 0, \
                            "there should be no invalid here", __file__)

        kepttest2duptest_map= {}
        test2keptdup = {}
        # get fdupes with metatests
        # TODO: check that each result of relpath is really a test
        for dup_tuple in dup_list:
            k_dir = KTestTestFormat.get_dir(dup_tuple[0], folders)
            key = os.path.relpath(dup_tuple[0], k_dir)
            key = DriversUtils.make_meta_element(key, \
                                                    folder2alias[k_dir])
            kepttest2duptest_map[key] = []
            for dp in dup_tuple[1:]:
                v_dir = KTestTestFormat.get_dir(dp, folders)
                val = os.path.relpath(dp, v_dir)
                val = DriversUtils.make_meta_element(val, \
                                                    folder2alias[v_dir])
                kepttest2duptest_map[key].append(val)
                test2keptdup[val] = key
        
        return kepttest2duptest_map, test2keptdup
    #~ def cross_folder_fdupes()
#~ class KTestTestFormat
